from moviepy.editor import VideoFileClip, CompositeVideoClip, ImageClip, clips_array
from typing import Union
from math import log as math_log, pow as math_pow

class MultipliedMoviepyEffect:
    """
    Generates a clips array with the provided 'clip' being shown
    'times' times (this parameter must be a pow of 4). This
    method has been created to be used internally with our own
    default methods.
    """
    __parameters = {}

    def __init__(self, clip: Union[VideoFileClip, CompositeVideoClip, ImageClip], times: int = 4):
        self.__clip = clip
        self.__parameters['times'] = times

    def __process_parameters(self):
        if not self.__parameters['times']:
            self.__parameters['times'] = 4

        if self.__parameters['times'] <= 1 or not math_log(self.__parameters['times'], 4).is_integer():
            # TODO: Raise exception instead of forcing 4 (?)
            self.__parameters['times'] = 4

        return self.__parameters
    
    def apply(self):
        """
        Applies the effect to the provided 'clip' and with the also
        provided parameters needed by this effect.
        """
        if not self.__clip:
            return None
        
        self.__process_parameters()
        
        audio = self.__clip.audio
        size = (self.__clip.w, self.__clip.h)

        # We will dynamically build the matrix
        row = []
        group = []
        # 4^3 = 64 => 8x8 = 2^3x2^3  and  4^2 = 16 => 4x4 = 2^2x2^2
        range_limit_value = math_pow(2, math_log(self.__parameters['times'], 4))
        for i in range(int(self.__parameters['times'] / range_limit_value)):
            row = []
            for j in range(int(self.__parameters['times'] / range_limit_value)):
                row.append(self.__clip)
            group.append(row)

        # When building 'clips_array' you sum the resolutions, so if you add four videos
        # of 1920x1080, you'll get one video of 4x(1920x1080) that will be impossible to
        # exported and unexpected. We resize it to avoid this and we don't resize each
        # clip before because they lose quality
        return clips_array(group).resize(size).set_audio(audio)


from faster_whisper import WhisperModel
# TODO: What if 'whisper_timestamped' is better? We should use it
import whisper_timestamped

def get_transcription_text(audio_filename, model = 'base'):
    """
    Receives an 'audio_filename' and gets the transcription of it,
    returning only the text. This is a simple and faster method
    to be used when we simply want the text without timestamps.
    """
    model = WhisperModel(model)

    segments, info = model.transcribe(audio_filename, language = 'es', beam_size = 5)

    text = ""
    for segment in segments:
        text += segment.text + ' '

    text = text.strip()

    return text

# Very interesting (https://github.com/m-bain/whisperX) to get timestamps of each word
# But the one I use is this one (https://github.com/linto-ai/whisper-timestamped)
def get_transcription_with_timestamps(audio_filename, model = 'small', initial_prompt = None):
    """
    Receives an 'audio_filename' and makes a transcription of it, returning a 
    list of 'segments' in the result.

    You can pass an 'initial_prompt' to help the model to recognize some specific
    words and to have a better performance.
    """
    audio = whisper_timestamped.load_audio(audio_filename)
    model = whisper_timestamped.load_model(model, device = "cpu")
    
    # I do this 'initial_prompt' formatting due to some issues when using it 
    # as it was. I read in Cookguide to pass natural sentences like this below
    # and it seems to be working well, so here it is:
    if initial_prompt != None:
        #initial_prompt = '""""' + initial_prompt + '""""'
        #initial_prompt = 'I know exactly what is said in this audio and I will give it to you (between double quotes) to give me the exact transcription. The audio says, exactly """"' + initial_prompt + '""""'
        initial_prompt = 'I will give you exactly what the audio says (the output), so please ensure it fits. The output must be """"' + initial_prompt + '""""'

    # 'vad' = True would remove silent parts to decrease hallucinations
    # 'detect_disfluences' detects corrections, repetitions, etc. so the
    # word prediction should be more accurate. Useful for natural narrations
    result = whisper_timestamped.transcribe(model, audio, language = "es", initial_prompt = initial_prompt)
    
    return result


"""
Flux is an image generation software created by Blackforest Labs
(https://blackforestlabs.ai/). They have 3 different models and
2 of those are free-to-use.

Those models ('dev' and 'schnell') have been implemented here to
be used to generate images.

The HugginFace endpoints I've found when creating this file:
- https://huggingface.co/black-forest-labs/FLUX.1-schnell?inference_api=true
- https://huggingface.co/black-forest-labs/FLUX.1-dev?inference_api=true
"""
from yta_multimedia.experimental.image.generation.ai.objects.huggingface import HugginFace
from typing import Union

SCHNELL_HUGGINFACE_ENDPOINT = 'https://api-inference.huggingface.co/models/black-forest-labs/FLUX.1-schnell'
DEV_HUGGINFACE_ENDPOINT = 'https://api-inference.huggingface.co/models/black-forest-labs/FLUX.1-dev'

def generate_dev(prompt: str, output_filename: Union[str, None] = None):
    """
    Generates an image using the Flux 'dev' model, through the 
    HugginFace endpoint, and returns it read as PIL image. If
    'output_filename' provided, it will be also stored locally
    with that name.

    @param
        **prompt**
        The prompt you want to use in the image generation. The
        more specific and explicit you are, the more accurate 
        result you get.

    @param
        **output_filename**
        The filename you want for the image to be stored locally.
        If None provided it won't be stored.
    """
    if not prompt:
        raise Exception('No valid "prompt" provided.')
    
    return HugginFace(DEV_HUGGINFACE_ENDPOINT).generate_image({'inputs': prompt}, output_filename)

def generate_schnell(prompt: str, output_filename: Union[str, None] = None):
    """
    Generates an image using the Flux 'dev' model, through the 
    HugginFace endpoint, and returns it read as PIL image. If
    'output_filename' provided, it will be also stored locally
    with that name.

    @param
        **prompt**
        The prompt you want to use in the image generation. The
        more specific and explicit you are, the more accurate 
        result you get.

    @param
        **output_filename**
        The filename you want for the image to be stored locally.
        If None provided it won't be stored.
    """
    if not prompt:
        raise Exception('No valid "prompt" provided.')
    
    return HugginFace(SCHNELL_HUGGINFACE_ENDPOINT).generate_image({'inputs': prompt}, output_filename)

# TODO: Try to download the models and implement with Diffusers
# https://huggingface.co/black-forest-labs/FLUX.1-schnell?inference_api=true
# https://huggingface.co/black-forest-labs/FLUX.1-schnell?inference_api=true

# Interesting: https://www.youtube.com/watch?v=P-rzgaIfZCo


from moviepy.editor import VideoFileClip, concatenate_videoclips
from yta_general_utils.type_checker import variable_is_type
from typing import Union

# TODO: Move this method to another site I think and refactor a little bit
# but it is working properly. I should receive a files list to merge 
# (maybe with abspath needed?).
def concatenate_videos_ffmpeg(abspath, output_filename = 'merged.mp4'):
    """
    This method concatenates the video files existing in the provided
    'abspath' and builds a new video with the name provided as
    'output_filename' in the same 'abspath' folder.

    This method uses the 'ffmpeg' to concatenate videos. Folder in
    'abspath' should contain only videos and with the same resolution.
    """
    # 1. List video files in folder
    import os
    from subprocess import run
    # Get the list of all files and directories
    files = os.listdir(abspath)
    files = [f for f in files if os.path.isfile(abspath + '/' + f)]

    # Write each file in a row with "file 'filename.ext'" format
    text = ''
    for file in files:
        text += "file '" + abspath + "/" + file + "'\n"

    filename = 'append_videos.txt'
    output_filename = abspath + '/' + output_filename
    with open(filename, 'w') as the_file:
        the_file.write(text)

    print(text)
    print(output_filename)

    command = 'ffmpeg -f concat -safe 0 -i ' + filename + ' -c copy ' + output_filename
    run(command)

def concatenate_moviepy(videoclips: list[VideoFileClip], output_filename: Union[str, None] = None):
    """
    Concatenates the provided moviepy 'videoclips' into a new one.

    If 'output_filename' provided, the new clip be stored with that
    name and that name.

    This method will return the new clips array.
    """
    if not videoclips:
        return None
    
    if variable_is_type(output_filename, str):
        if not output_filename:
            return None

    clip = concatenate_videoclips(videoclips)

    if output_filename:
        clip.write_videofile(output_filename)

    return clip


from yta_general_utils.file_downloader import get_id_from_google_drive_url
from yta_general_utils.tmp_processor import create_tmp_filename
from yta_multimedia.greenscreen.utils import get_greenscreen_area_details
from yta_general_utils.type_checker import is_google_drive_url
from yta_multimedia.greenscreen.classes.greenscreen_details import GreenscreenDetails
from yta_multimedia.greenscreen.classes.greenscreen_area_details import GreenscreenAreaDetails
from yta_multimedia.video.frames import get_frame_from_video_by_frame_number
from yta_multimedia.greenscreen.enum.greenscreen_type import GreenscreenType
from yta_multimedia.greenscreen.utils import GREENSCREENS_FOLDER
from yta_multimedia.resources import get_resource

def get_greenscreen_details(greenscreen_filename_or_google_drive_url: str, type: GreenscreenType):
    """
    Method to obtain greenscreen area and details that must be
    used by ImageGreenscreen and VideoGreenscreen to automatically
    detect greenscreens in their resources.
    """
    if not greenscreen_filename_or_google_drive_url:
        return None
    
    # TODO: Check that 'type' is GreenscreenType type

    # We will need the resource filename or google drive url and
    # the image to extract the data from
    RESOURCE_FILENAME = greenscreen_filename_or_google_drive_url
    TMP_FILENAME = greenscreen_filename_or_google_drive_url

    if type == GreenscreenType.IMAGE:
        # We have the final resource and the image to extract data
        if is_google_drive_url(greenscreen_filename_or_google_drive_url):
            google_drive_id = get_id_from_google_drive_url(greenscreen_filename_or_google_drive_url)
            filename = GREENSCREENS_FOLDER + google_drive_id + '/greenscreen.png'
            TMP_FILENAME = get_resource(greenscreen_filename_or_google_drive_url, filename)
    else: # GreenscreenType.VIDEO
        if is_google_drive_url(greenscreen_filename_or_google_drive_url):
            google_drive_id = get_id_from_google_drive_url(greenscreen_filename_or_google_drive_url)
            filename = GREENSCREENS_FOLDER + google_drive_id + '/greenscreen.mp4'
            RESOURCE_FILENAME = get_resource(greenscreen_filename_or_google_drive_url, filename)
            TMP_FILENAME = create_tmp_filename('tmp_gs_autodetect.png')
            get_frame_from_video_by_frame_number(RESOURCE_FILENAME, 0, TMP_FILENAME)

    # TODO: Autodetect and instantiate a Greenscreen
    details = get_greenscreen_area_details(TMP_FILENAME)

    if not details['rgb_color']:
        raise Exception('No automatic greenscreen detected in "' + greenscreen_filename_or_google_drive_url + '". Aborting.')

    greenscreen_filename_or_google_drive_url = GreenscreenDetails(
        greenscreen_areas = [
            GreenscreenAreaDetails(
                rgb_color = details['rgb_color'],
                upper_left_pixel = details['upper_left_pixel'],
                lower_right_pixel = details['lower_right_pixel'],
                frames = None # TODO: Implement a way of handling frames
            )
        ],
        filename_or_google_drive_url = RESOURCE_FILENAME,
        type = type
    )

    return greenscreen_filename_or_google_drive_url


from PIL import Image

def pixelate_image_file(image_filename: str, i_size, output_filename: str):
    """
    Pixelates the provided 'image_filename' and saves it as the 'output_filename'.
    The 'i_size' is the pixelating square. The smaller it is, the less pixelated 
    its.

    'i_size' must be a tuple such as (8, 8) or (16, 16).
    """
    if not image_filename:
        return None
    
    # TODO: Handle 'i_size' format and check
    
    if not output_filename:
        return None

    img = Image.open(image_filename)

    # Convert to small image
    small_img = img.resize(i_size,Image.BILINEAR)

    # Resize to output size
    res = small_img.resize(img.size, Image.NEAREST)

    res.save(output_filename)


"""
Here we are using the https://message.style/app/editor
platform with a Web Scrapper to obtain the nice Discord
images they generate.
"""
from selenium.webdriver.common.keys import Keys
from yta_general_utils.web.scrapper.chrome_scrapper import ChromeScrapper
from PIL import Image
from time import sleep

import io
import json
import datetime


WEB_URL = 'https://message.style/app/editor'

class SingletonMeta(type):
    """
    The Singleton class can be implemented in different ways in Python. Some
    possible methods include: base class, decorator, metaclass. We will use the
    metaclass because it is best suited for this purpose.
    """

    _instances = {}

    def __call__(cls, *args, **kwargs):
        """
        Possible changes to the value of the `__init__` argument do not affect
        the returned instance.
        """
        if cls not in cls._instances:
            instance = super().__call__(*args, **kwargs)
            cls._instances[cls] = instance
        return cls._instances[cls]

class DiscordMessageImageGenerator(metaclass = SingletonMeta):
    """
    Singleton class to generate Discord Message images.
    """
    def __init__(self):
        if not hasattr(self, 'scrapper'):
            self.scrapper = ChromeScrapper(True)

    def __build_string_json(self, username: str, user_avatar_url: str, message: str, image_url: str = None):
        """
        Returns the json that will create the message in the platform,
        but as a string.
        """
        # TODO: Sanitize this better, please
        # Check this: https://forum.uipath.com/t/how-to-remove-r-n-from-this-json/747396/11

        embeds = []
        if image_url:  # if image
            embeds.append({
                "id": 652627557,
                "color": 3289650,
                "fields": [],
                "image": {
                    "url": image_url
                }
            })

        json_content = {
            "content": message,
            "tts": False,
            "username": username,
            "avatar_url": user_avatar_url,
            "embeds": embeds,
            "components": [],
            "actions": {}
        }

        return json.dumps(json_content, ensure_ascii = False).encode('utf-8').decode('utf-8')

    def generate(self, username: str, user_avatar_url: str, message: str, badge_text: str = None, datetime_text: str = None, image_url: str = None, output_filename: str = None):
        """
        Generates a Discord message with the provided parameters and
        stores it locally as 'output_filename'.

        If you want to do a line break you must use "\\ \\n" but 
        toghether, without the blank space I left and without the
        quotes.

        If 'badge_text' is None the badge will be not shown. If 
        'datetime_text' is None, it will display the current datetime
        as 'Hoy a las XX:XX'.
        """
        if not username:
            raise Exception('No "username" provided')
        
        if not user_avatar_url:
            raise Exception('No "user_avatar_url" provided.')
        
        if not message:
            message = ''
        
        # We go to the main website
        self.scrapper.go_to_web_and_wait_util_loaded(WEB_URL)

        # TODO: Check if provided 'image_url' is a valid image
        json_string = self.__build_string_json(username, user_avatar_url, message, image_url)
        # We need to manually add the string to clipboard to 
        # be able to paste it
        self.scrapper.add_to_clipboard(json_string)

        # We will now change the JSON to format the message we want
        self.scrapper.go_to_web_and_wait_util_loaded(WEB_URL + '/json')
        
        # Focus the editor to be able to paste
        self.scrapper.press_key_x_times(Keys.TAB, 25) # Manually detected, to focus on json editor

        # Select the whole previous json, delete it and paste our own
        self.scrapper.press_ctrl_a()
        self.scrapper.press_key_x_times(Keys.DELETE, 1)
        active_element = self.scrapper.driver.switch_to.active_element
        active_element.click()
        self.scrapper.press_key_x_times(Keys.DELETE, 1)
        self.scrapper.press_ctrl_v_on_element(active_element)
        # Press the save button
        save_button = self.scrapper.find_element_by_text('button', 'Save')
        save_button.click()
        # TODO: This could be a 'wait-for-element-change' waiting,
        # by now the amount of seconds is manually chosen
        sleep(2)
        message_element = self.scrapper.find_element_by_class('div', 'discord-message')

        # Hide bot badge
        if not badge_text:
            self.scrapper.execute_script('document.getElementsByClassName("discord-application-tag")[0].style.display = "none";')
        else:
            self.scrapper.execute_script('document.getElementsByClassName("discord-application-tag")[0].innerHTML = "' + badge_text + '";')

        # Set datetime
        if not datetime_text:
            now = datetime.datetime.now().strftime("%H:%M")
            if now.startswith('0'):
                now = now[1:]
            datetime_text = 'hoy a las ' +  now
        self.scrapper.execute_script('document.getElementsByClassName("discord-message-timestamp pl-1")[0].innerHTML = "' + datetime_text + '";')
        
        # We return the screenshot or we store it and return the image filename
        screenshot = self.scrapper.screenshot_element(message_element, output_filename)

        if output_filename:
            return screenshot

        # TODO: Write this in docummentation
        # TODO: If unneeded, replace the dict in ChromeScrapper
        # for only the 'data' as we don't need to know the size
        return Image.open(io.BytesIO(screenshot['data']))


"""
Useful discord messages below

{
  "content": "Welcome to **Embed Generator**! 🎉 Create stunning embed messages for your Discord server with ease!\n\nIf you're ready to start, simply click on the \"Clear\" button at the top of the editor and create your own message.\n\nShould you need any assistance or have questions, feel free to join our [support server](/discord) where you can connect with our helpful community members and get the support you need.\n\nWe also have a [complementary bot](/invite) that enhances the experience with Embed Generator. Check out our [Discord bot](/invite) which offers features like formatting for mentions, channels, and emoji, creating reaction roles, interactive components, and more.\n\nLet your creativity shine and make your server stand out with Embed Generator! ✨",
  "tts": false,
  "embeds": [
    {
      "id": 652627557,
      "title": "About Embed Generator",
      "description": "Embed Generator is a powerful tool that enables you to create visually appealing and interactive embed messages for your Discord server. With the use of webhooks, Embed Generator allows you to customize the appearance of your messages and make them more engaging.\n\nTo get started, all you need is a webhook URL, which can be obtained from the 'Integrations' tab in your server's settings. If you encounter any issues while setting up a webhook, our bot can assist you in creating one.\n\nInstead of using webhooks you can also select a server and channel directly here on the website. The bot will automatically create a webhook for you and use it to send the message.",
      "color": 2326507,
      "fields": []
    },
    {
      "id": 10674342,
      "title": "Discord Bot Integration",
      "description": "Embed Generator offers a Discord bot integration that can further enhance your the functionality. While it is not mandatory for sending messages, having the bot on your server gives you access to a lot more features!\n\nHere are some key features of our bot:",
      "color": 2326507,
      "fields": [
        {
          "id": 472281785,
          "name": "Interactive Components",
          "value": "With our bot on your server you can add interactive components like buttons and select menus to your messages. Just invite the bot to your server, select the right server here on the website and you are ready to go!"
        },
        {
          "id": 608893643,
          "name": "Special Formatting for Mentions, Channels, and Emoji",
          "value": "With the /format command, our bot provides special formatting options for mentions, channel tags, and ready-to-use emoji. No more manual formatting errors! Simply copy and paste the formatted text into the editor."
        },
        {
          "id": 724530251,
          "name": "Recover Embed Generator Messages",
          "value": "If you ever need to retrieve a previously sent message created with Embed Generator, our bot can assist you. Right-click or long-press any message in your server, navigate to the apps menu, and select Restore to Embed Generator. You'll receive a link that leads to the editor page with the selected message."
        },
        {
          "id": 927221233,
          "name": "Additional Features",
          "value": "Our bot also supports fetching images from profile pictures or emojis, webhook management, and more. Invite the bot to your server and use the /help command to explore all the available features!"
        }
      ]
    }
  ],
  "components": [],
  "actions": {}
}
"""


from yta_multimedia.image.edition.filter.workers.sketch.linedraw import apply_line_sketch_to_image
from yta_multimedia.greenscreen.utils import numpy_to_pil
from PIL import Image
from typing import Union

import cv2
import numpy as np

def image_to_sketch(image: str, output_filename: Union[str, None] = None):
    """
    Turns the provided 'image' into an sketch that is made 
    by white and black colors.
    """
    # Thank you: https://github.com/code-kudos/Convert-any-image-to-sketch-with-python/blob/main/Sketch.py
    # TODO: Image can be an array containing the image (numpy.ndarray)
    # TODO: Check if image provided is # numpy.ndarray or PIL image
    # if not image:
    #     return None

    # TODO: Check 'output_filename'
    # if isinstance(image, np.ndarray):
    #     image = numpy_to_pil(image)
    # else:
    #     # TODO: Check if image is actually a Image.opened(image)
    #     image = Image.open(image)

    # numpy array is working here
    grey_img = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)
    invert_img = cv2.bitwise_not(grey_img)
    blur_img = cv2.GaussianBlur(invert_img, (277, 277), 0)
    invblur_img = cv2.bitwise_not(blur_img)
    sketch_img = cv2.divide(grey_img, invblur_img, scale = 256.0)
    rgb_sketch = cv2.cvtColor(sketch_img, cv2.COLOR_BGR2RGB)

    sketch_image = Image.fromarray(rgb_sketch)

    if output_filename:
        sketch_image.save(output_filename)

    return sketch_image

def image_to_line_sketch(image: str, output_filename: Union[str, None] = None):
    """
    Turns the provided 'image' into an line sketch that is made 
    by white and black colors.
    """
    # Thank you: https://github.com/LingDong-/linedraw
    # TODO: Image can be an array containing the image (numpy.ndarray)
    # TODO: Check if image provided is # numpy.ndarray or PIL image
    # if not image:
    #     return None

    # Code here below:
    # TODO: I think the 'output_filename' here is not needed
    sketch_image = apply_line_sketch_to_image(image, output_filename)

    if output_filename:
        sketch_image.save(output_filename)

    return sketch_image


from yta_multimedia.greenscreen.utils import pil_to_numpy
from yta_general_utils.type_checker import variable_is_type
from yta_general_utils.file_processor import file_is_video_file
from yta_multimedia.video.frames import get_all_frames_from_video
from yta_multimedia.image.edition.filter.sketch import image_to_sketch, image_to_line_sketch
from moviepy.editor import VideoFileClip, CompositeVideoClip, ImageClip, ImageSequenceClip
from typing import Union
from types import FunctionType

def video_to_sketch_video(video: Union[str, VideoFileClip, CompositeVideoClip, ImageClip], output_filename: Union[str, None]):
    # TODO: Document it
    return __video_to_frame_by_frame_filtered_video(video, image_to_sketch, output_filename)

def video_to_line_sketch_video(video: Union[str, VideoFileClip, CompositeVideoClip, ImageClip], output_filename: Union[str, None]):
    """
    This method is very very slow. I should try to optimize it or just
    use not, because it doesn't make sense as a video.
    """
    # TODO: Document it
    return __video_to_frame_by_frame_filtered_video(video, image_to_line_sketch, output_filename)

def __video_to_frame_by_frame_filtered_video(video: Union[str, VideoFileClip, CompositeVideoClip, ImageClip], filter_func: FunctionType, output_filename: Union[str, None] = None):
    """
    Internal function to be used by any of our video editing methods
    that actually use image filter frame by frame. They do the same
    by only changing the filter we apply.
    """
    # TODO: Check if 'filter_func' is a function

    if not video:
        raise Exception('No "video" provided.')
    
    if variable_is_type(video, str):
        if not file_is_video_file(video):
            return None
        
        video = VideoFileClip(video)

    original_frames = get_all_frames_from_video(video)
    sketched_frames = []
    for original_frame in original_frames:
        sketched_frames.append(pil_to_numpy(filter_func(original_frame)))

    sketched_video = ImageSequenceClip(sketched_frames, fps = video.fps).set_audio(video.audio)

    if output_filename:
        sketched_video.write_videofile(output_filename)

    return sketched_video




from moviepy.editor import vfx, VideoFileClip, CompositeVideoClip, ImageClip
from typing import Union

class BlackAndWhiteMoviepyEffect:
    """
    This effect will make the clip appear in black and
    white colors.
    """
    __MOVIEPY_EFFECT_NAME = 'blackwhite'
    __parameters = {}

    def __init__(self, clip: Union[VideoFileClip, CompositeVideoClip, ImageClip]):
        self.__clip = clip

    def __get_moviepy_vfx_effect(self):
        return getattr(vfx, self.__MOVIEPY_EFFECT_NAME, None)
    
    def __process_parameters(self):
        return self.__parameters
    
    def apply(self):
        """
        Applies the effect to the provided 'clip' and with the also
        provided parameters needed by this effect.
        """
        return self.__clip.fx(self.__get_moviepy_vfx_effect(), **self.__process_parameters())



import librosa

def get_song_tempo_and_beats(filename: str):
    """
    Analyzes the provided song 'filename' and returns a dict that
    contains the song estimated 'tempo' and 'beats' time moments
    (in seconds).

    @param
        **filename**
        The filename of a file that contains a song to be analyzed.
    """
    # Thanks: https://dev.to/highcenburg/getting-the-tempo-of-a-song-using-librosa-4e5b
    audio_file = librosa.load(filename)
    y, sr = audio_file
    estimated_tempo, beat_frames = librosa.beat.beat_track(y = y, sr = sr)
    beat_times = librosa.frames_to_time(beat_frames, sr = sr)

    response = {
        'tempo': estimated_tempo,
        'beats': beat_times
    }

    return response


from moviepy.editor import VideoFileClip
from typing import Union
from yta_general_utils.file_processor import file_is_video_file
from yta_general_utils.type_checker import variable_is_type
from scenedetect import SceneManager, open_video, ContentDetector

def detect_scenes_from_video(video_input: Union[VideoFileClip, str]):
    """
    This method detects the different scenes that are in the provided 'video_input'
    video file and returns an array containing 'start' and 'end' elements. Each of 
    those contains a 'second' and 'frame' field.

    TODO: This link https://www.scenedetect.com/docs/latest/api.html#example says that
    there is a backend 'VideoStreamMoviePy' class to process it with moviepy. This is
    interesting to use in a clip, and not in a filename.
    """
    # This comes from here: https://www.scenedetect.com/
    # Other project: https://github.com/slhck/scenecut-extractor (only ffmpeg)
    if not video_input:
        return None
    
    if variable_is_type(video_input, str):
        if not file_is_video_file(video_input):
            return None
        
        video_input = VideoFileClip(video_input)

    # TODO: Check if thiss 'filename' includes the abspath
    video = open_video(video_input.filename, backend = 'moviepy')
    scene_manager = SceneManager()
    scene_manager.add_detector(ContentDetector(threshold = 20))  # 27 is recommended
    scene_manager.detect_scenes(video)

    scenes = []
    for scene in scene_manager.get_scene_list():
        scenes.append({
            'start': {
                'second': scene[0].get_seconds(),
                'frame': scene[0].get_frames()
            },
            'end': {
                'second': scene[1].get_seconds(),
                'frame': scene[1].get_frames()
            }
        })

    return scenes

    # This code below is working but it is quite old
    
    from scenedetect import VideoManager, SceneManager, StatsManager
    from scenedetect.detectors import ContentDetector
    from scenedetect.scene_manager import save_images, write_scene_list_html

    video_manager = VideoManager([video_filename])
    stats_manager = StatsManager()
    scene_manager = SceneManager(stats_manager)

    scene_manager.add_detector(ContentDetector(threshold = 30))
    video_manager.set_downscale_factor()

    video_manager.start()
    scene_manager.detect_scenes(frame_source = video_manager)

    scenes = scene_manager.get_scene_list()
    print(f'{len(scenes)} scenes detected!')

    save_images(
        scenes,
        video_manager,
        num_images = 1,
        image_name_template = '$SCENE_NUMBER',
        output_dir = 'scenes')

    for scene in scenes:
        start, end = scene

        # your code
        print(f'{start.get_seconds()} - {end.get_seconds()}')

    return scenes


# This is to work with abruptness and sooness (https://moviepy.readthedocs.io/en/latest/ref/videofx/moviepy.video.fx.all.accel_decel.html)
    
# I previously had 4.4.2 decorator for moviepy. I forced 4.0.2 and it is apparently working

"""
Interesting:
- https://www.youtube.com/watch?v=Ex1kuxe6jRg (el canal entero tiene buena pinta)
- https://www.youtube.com/watch?v=m6chqKlhpPo Echarle un vistazo a ese tutorial
- https://zulko.github.io/moviepy/ref/videofx.html
- https://stackoverflow.com/questions/48491070/how-to-flip-an-mp4-video-horizontally-in-python
"""



TYPING_KEYBOARD_3_SECONDS_GOOGLE_DRIVE_DOWNLOAD_URL = 'https://drive.google.com/file/d/179wzrnFwcy15wG4X1_BcGuN6JUSYE6ej/view?usp=sharing'
SILENCE_10_SECONDS_GOOGLE_DRIVE_DOWNLOAD_URL = 'https://drive.google.com/file/d/1yL8eAMS5FlT3L3tjJsftcema5BW-HFcJ/view?usp=sharing'


from PIL import Image
from transformers import BlipProcessor, BlipForConditionalGeneration
from yta_general_utils.type_checker import variable_is_type
from typing import Union

def describe_image(image_filename: Union[str, Image.Image]):
    """
    Describes the provided 'image_filename' by using the Salesforce
    blip image captioning system.
    """
    if not image_filename:
        return None
    
    if variable_is_type(image_filename, str):
        image_filename = Image.open(image_filename)

    # models are stored in C:\Users\USERNAME\.cache\huggingface\hub
    processor = BlipProcessor.from_pretrained('Salesforce/blip-image-captioning-base')
    model = BlipForConditionalGeneration.from_pretrained('Salesforce/blip-image-captioning-base')
    inputs = processor(image_filename, return_tensors = 'pt')
    out = model.generate(**inputs)
    description = processor.decode(out[0], skip_special_tokes = True)

    return description


from typing import Union
from moviepy.editor import ImageClip
from yta_general_utils.type_checker import variable_is_type

def generate_videoclip_from_image(image_filename: Union[ImageClip, str], duration: float = 1, output_filename: Union[str, None] = None):
    """
    Receives an image as 'image_filename' and creates an ImageClip of
    'duration' seconds. It will be also stored as a file if 
    'output_filename' is provided.

    # TODO: Should this method go into 'video.utils' instead of here (?)
    """
    if not image_filename:
        return None
    
    if duration <= 0:
        return None
    
    if not duration:
        return None
    
    if variable_is_type(output_filename, str):
        if not output_filename:
            return None
    
    if variable_is_type(image_filename, str):
        # ADV: By now we are limiting this to 60 fps
        image_filename = ImageClip(image_filename).set_fps(60).set_duration(duration)

    if output_filename:
        image_filename.write_videofile(output_filename)

    return image_filename


from yta_multimedia.greenscreen.classes.greenscreen_details import GreenscreenDetails
from yta_general_utils.type_checker import variable_is_type
from moviepy.editor import ImageClip
from typing import Union
from yta_multimedia.greenscreen.classes.greenscreen_details import GreenscreenDetails
from moviepy.editor import VideoFileClip, ImageClip, CompositeVideoClip, vfx, concatenate_videoclips
from yta_general_utils.file_processor import file_is_video_file
from yta_multimedia.greenscreen.custom.utils import get_greenscreen_details
from yta_multimedia.greenscreen.enum.greenscreen_type import GreenscreenType

class VideoGreenscreen:
    greenscreen: GreenscreenDetails = None

    def __init__(self, greenscreen: Union[GreenscreenDetails, str]):
        if variable_is_type(greenscreen, str):
            # We need to automatically detect greenscreen details
            greenscreen = get_greenscreen_details(greenscreen, GreenscreenType.VIDEO)

        self.greenscreen = greenscreen
       
    def from_image_to_image(self):
        # TODO: Does this make sense? We are working with a video, so
        # trying to generate just a frame, an image, does make sense?
        # TODO: By the way, this is not working yet
        pass

    def from_image_to_video(self, image_filename: str, duration: float = 3.0, output_filename: Union[str, None] = None):
        """
        Receives an 'image_filename', places it into the greenscreen and generates
        a video of 'duration' seconds of duration that is returned. This method
        will store locally the video if 'output_filename' is provided.
        """
        if not image_filename:
            return None
        
        if not duration:
            return None
        
        if not output_filename:
            return None
        
        imageclip = ImageClip(image_filename, duration = duration)
        final_clip = self.from_video_to_video(imageclip)

        if output_filename:
            final_clip.write_videofile(output_filename)

        return final_clip
    
    def from_video_to_image(self):
        # TODO: Does this make sense? By now it is not implemented
        pass

    def from_video_to_video(self, video: Union[str, VideoFileClip, CompositeVideoClip, ImageClip], output_filename: Union[str, None] = None):
        """
        Inserts the provided 'video' in the greenscreen and returns the
        CompositeVideoClip that has been created. If 'output_filename' 
        provided, it will be written locally with that file name.

        The provided 'video' can be a filename or a moviepy video clip.

        TODO: Careful when video is longer than greenscreen
        """
        if not video:
            return None
        
        if variable_is_type(video, str):
            if not file_is_video_file(video):
                return None
            
            video = VideoFileClip(video)

        # TODO: Change this 
        # We could have different areas in which append 1 or more 
        # videos, so it would be handled different. By now I'm
        # considering we only have one

        TMP_FILENAME = self.greenscreen.get_filename()

        # We choose the first greenscreen area to work with
        greenscreen_area = self.greenscreen.greenscreen_areas[0]
        
        # Create the clip
        green_screen_clip = VideoFileClip(TMP_FILENAME).fx(vfx.mask_color, color = greenscreen_area.rgb_color, thr = 100, s = 5)

        width = greenscreen_area.lower_right_pixel[0] - greenscreen_area.upper_left_pixel[0]
        # If the provided clip is shorter than our green screen: easy, crop the green screen
        # If the provided clip is longer than our green screen: I use the green screen duration
        # and let the clip be original the rest of the time
        # TODO: Improve resizing process as it is failing...
        if green_screen_clip.duration > video.duration:
            green_screen_clip = green_screen_clip.set_duration(video.duration)
            # Clip will be displayed inside the green screen area
            video = video.resize(width = width).set_position((greenscreen_area.upper_left_pixel[0], greenscreen_area.upper_left_pixel[1]))
            video = CompositeVideoClip([video, green_screen_clip], size = green_screen_clip.size)
        elif video.duration > green_screen_clip.duration:
            # First subclip will be displayed inside the green screen area
            first_clip = video.subclip(0, green_screen_clip.duration).resize(width = width).set_position((greenscreen_area.upper_left_pixel[0], greenscreen_area.upper_left_pixel[1]))
            # Second clip will be as the original one
            second_clip = video.subclip(green_screen_clip.duration, video.duration)
            video = concatenate_videoclips([
                CompositeVideoClip([first_clip, green_screen_clip], size = green_screen_clip.size),
                second_clip
            ])
        else:
            video = CompositeVideoClip([video, green_screen_clip], size = green_screen_clip.size)

        if output_filename:
            video.write_videofile(output_filename, fps = video.fps)

        return video


from yta_multimedia.greenscreen.classes.greenscreen_details import GreenscreenDetails
from yta_general_utils.type_checker import variable_is_type
from moviepy.editor import ImageClip
from typing import Union
from yta_multimedia.greenscreen.classes.greenscreen_details import GreenscreenDetails
from moviepy.editor import VideoFileClip, ImageClip, CompositeVideoClip, vfx
from yta_general_utils.file_processor import file_is_video_file
from yta_multimedia.resources import get_resource
from yta_multimedia.greenscreen.custom.utils import get_greenscreen_details
from yta_general_utils.file_downloader import get_id_from_google_drive_url
from yta_multimedia.greenscreen.enum.greenscreen_type import GreenscreenType
from yta_multimedia.greenscreen.utils import GREENSCREENS_FOLDER
from PIL import Image, ImageDraw
from typing import Union, Any

import numpy as np

class ImageGreenscreen:
    greenscreen: GreenscreenDetails = None

    def __init__(self, greenscreen: Union[GreenscreenDetails, str]):
        if variable_is_type(greenscreen, str):
            # We need to automatically detect greenscreen details
            greenscreen = get_greenscreen_details(greenscreen, GreenscreenType.IMAGE)

        self.greenscreen = greenscreen

    def __process_elements_and_save(self, output_filename):
        """
        Processes the greenscreen by writing the title, description
        and any other available element, and stores it locally as
        'output_filename' once processed.
        """
        base = Image.open(self.greenscreen.filename_or_google_drive_url)
        draw = ImageDraw.Draw(base)

        # TODO: I preserve this code for the future
        # # We need to write title if existing
        # if self.__title:
        #     title_position = (self.__title_x, self.__title_y)
        #     draw.text(title_position, self.__title, font = self.__title_font, fill = self.__title_color)

        # if self.__description:
        #     description_position = (self.__description_x, self.__description_y)
        #     draw.text(description_position, self.__description, font = self.__description_font, fill = self.__description_color)

        # TODO: Handle anything else here

        # We save the image
        base.save(output_filename, quality = 100)

    # TODO: Make this below expect a type 'Image', but 'Image' or
    # 'Image.Image' returns each arg must be a type. Got <module 
    # 'PIL.Image'
    def from_image_to_image(self, image: Union[str, Any], output_filename: Union[str, None] = None):
        """
        Receives an 'image_filename', places it into the greenscreen and generates
        another image that is stored locally as 'output_filename'.
        """
        # TODO: This is failing when numpy parameter because
        # The truth value of an array with more than one 
        # element is ambiguous. Use a.any() or a.all()
        # if image == None:
        #     return None
        
        if not output_filename:
            return None
        
        if variable_is_type(image, str):
            # TODO: Check if image is valid

            image = np.asarray(Image.open(image))
        
        # Put the image into the greenscreen
        imageclip = ImageClip(image, duration = 1 / 60)
        final_clip = self.from_video_to_video(imageclip)

        # Generate the result image
        if output_filename:
            final_clip.save_frame(output_filename, t = 0)

        return final_clip.get_frame(t = 0)

    def from_image_to_video(self, image: str, duration: float = 3.0, output_filename: Union[str, None] = None):
        """
        Receives an 'image', places it into the greenscreen and generates
        a video of 'duration' seconds of duration that is returned. This method
        will store locally the video if 'output_filename' is provided.
        """
        if not image:
            return None
        
        if not duration:
            return None
        
        if not output_filename:
            return None
        
        if variable_is_type(image, str):
            # TODO: Check if image is valid

            image = np.asarray(Image.open(image))

        imageclip = ImageClip(image, duration = duration)
        final_clip = self.from_video_to_video(imageclip)

        if output_filename:
            final_clip.write_videofile(output_filename)

        return final_clip
    
    def from_video_to_image(self):
        # TODO: Does this make sense? By now it is not implemented
        pass

    def from_video_to_video(self, video: Union[str, VideoFileClip, CompositeVideoClip, ImageClip], output_filename: Union[str, None] = None):
        """
        Inserts the provided 'video' in the greenscreen and returns the
        CompositeVideoClip that has been created. If 'output_filename' 
        provided, it will be written locally with that file name.

        The provided 'video' can be a filename or a moviepy video clip.

        TODO: Careful when video is longer than greenscreen
        """
        if not video:
            return None
        
        if variable_is_type(video, str):
            if not file_is_video_file(video):
                return None
            
            video = VideoFileClip(video)

        # TODO: Change this 
        # We could have different areas in which append 1 or more 
        # videos, so it would be handled different. By now I'm
        # considering we only have one
        
        # We download the file to be able to use it
        # TODO: This should be refactor with new fields in ImageGreenscreen
        # such as 'filename' and 'google_drive_url' by separate
        # TODO: We should check if this is a filename or a Google Drive url
        # and act in consecuence
        #TMP_FILENAME = get_resource(self.greenscreen.filename_or_google_drive_url, self.greenscreen.filename_or_google_drive_url)
        google_drive_id = get_id_from_google_drive_url(self.greenscreen.filename_or_google_drive_url)
        TMP_FILENAME = get_resource(self.greenscreen.filename_or_google_drive_url, GREENSCREENS_FOLDER + google_drive_id + '/greenscreen.mp4')
        #TMP_FILENAME = download_file_from_google_drive(self.greenscreen.filename_or_google_drive_url, moving_greenscreen_frames_folder_relpath)

        # We process greenscreen elements if needed
        # TODO: This will be done in the future (writing titles)
        #tmp_filename = create_tmp_filename('tmp_gs.png')
        #self.__process_elements_and_save(tmp_filename)

        # We choose the first greenscreen area to work with
        greenscreen_area = self.greenscreen.greenscreen_areas[0]

        # We work to maintain the proportion
        ulx = greenscreen_area.upper_left_pixel[0]
        uly = greenscreen_area.upper_left_pixel[1]
        lrx = greenscreen_area.lower_right_pixel[0]
        lry = greenscreen_area.lower_right_pixel[1]
        
        # TODO: Please, review this below
        gs_width = lrx - ulx
        gs_height = lry - uly
        aspect_ratio = gs_width / gs_height
        if aspect_ratio > (1920 / 1080):
            # We keep the width, but we calculate the height to keep 16/9 aspect ratio
            gs_height_for_16_9 = gs_width * (1080 / 1920)
            # Now we know the new height, so we only have to center the video on 'y' axis
            difference = gs_height_for_16_9 - gs_height
            # My video (16/9) will not fit the whole width
            uly -= difference / 2
            lry += difference / 2
        elif aspect_ratio < (1920 / 1080):
            # We keep the height, but we calculate the width to keep 16/9 aspect ratio
            gs_width_for_16_9 = gs_height * (1920 / 1080)
            # Now we know the new width, so we only have to center the video on 'x' axis
            difference = gs_width_for_16_9 - gs_width
            ulx -= difference / 2
            lrx += difference / 2
        # TODO: Are we sure that those coords are int and make a real 16/9 with int numbers?
        
        # Create the clip
        green_screen_clip = ImageClip(TMP_FILENAME, duration = video.duration).fx(vfx.mask_color, color = greenscreen_area.rgb_color, thr = 100, s = 5)
        width = lrx - ulx
        video = video.resize(width = width).set_position((ulx, uly))

        final_clip = CompositeVideoClip([video, green_screen_clip], size = green_screen_clip.size)

        if output_filename:
            final_clip.write_videofile(output_filename, fps = video.fps)

        return final_clip


from utils.downloader.image import download_by_url
from utils.scrapper.chrome import start_chrome, go_to_and_wait_loaded
from selenium.webdriver.common.by import By
from selenium.webdriver.common.action_chains import ActionChains
from selenium.webdriver.common.keys import Keys
from objects.google_image_scraper import GoogleImageScraper

import urllib.request
import urllib
import urllib.parse
import imghdr
import posixpath
import re
import time
from itertools import filterfalse

IMAGE_FILE_EXTENSIONS = ['jpe', 'jpeg', 'jfif', 'exif', 'tiff', 'gif', 'bmp', 'png', 'webp', 'jpg']

class Bing:
    # Based on this (https://github.com/gurugaurav/bing_image_downloader/blob/master/bing_image_downloader/bing.py), but modified to my own purpose
    __instance = None

    def __new__(cls, ignore_repeated = True):
        if not Bing.__instance:
            Bing.__instance = object.__new__(cls)
        
        return Bing.__instance

    def __init__(self, output_dir, adult = '', timeout = 100,  filter = '', ignore_repeated = True):
        self.download_count = 0
        self.output_dir = output_dir
        self.adult = adult
        self.filter = filter
        self.ignore_repeated = ignore_repeated
        self.seen = set()
        self.downloaded_urls = set()

        self.timeout = timeout

        self.page_counter = 0
        self.headers = {
            'User-Agent': 'Mozilla/5.0 (X11; Linux x86_64) ' 
            'AppleWebKit/537.11 (KHTML, like Gecko) '
            'Chrome/23.0.1271.64 Safari/537.11',
            'Accept': 'text/html,application/xhtml+xml,application/xml;q=0.9,*/*;q=0.8',
            'Accept-Charset': 'ISO-8859-1,utf-8;q=0.7,*;q=0.3',
            'Accept-Encoding': 'none',
            'Accept-Language': 'en-US,en;q=0.8',
            'Connection': 'keep-alive'
        }

    def activate_ignore_repeated(self):
        self.ignore_repeated = True

    def deactivate_ignore_repeated(self):
        self.ignore_repeated = False

    def __get_filter(self, shorthand):
            if shorthand == "line" or shorthand == "linedrawing":
                return "+filterui:photo-linedrawing"
            elif shorthand == "photo":
                return "+filterui:photo-photo"
            elif shorthand == "clipart":
                return "+filterui:photo-clipart"
            elif shorthand == "gif" or shorthand == "animatedgif":
                return "+filterui:photo-animatedgif"
            elif shorthand == "transparent":
                return "+filterui:photo-transparent"
            else:
                return ""
            
    def __get_url(self, query, limit):
        """
        Returns the url to search the provided 'query' to get 'limit' images.
        """
        query = urllib.parse.quote_plus(query)

        # This below works
        #return 'https://www.bing.com/images/async?q=' + urllib.parse.quote_plus(query) + '&first=' + str(self.page_counter) + '&count=' + str(limit) + '&adlt=' + self.adult + '&qft=' + ('' if self.filter is None else self.__get_filter(self.filter))

        #url = 'https://www.google.com/search?q=' + query + '&sca_esv=ff48bec18c46a1e3&udm=2&biw=1368&bih=785&sxsrf=ACQVn0-wIFIFwKpRcfa_i7u23WlweS5fMg%3A1711107979098&ei=i2_9ZZrBBYmPxc8P2d2o8Aw&ved=0ahUKEwjaspD65YeFAxWJR_EDHdkuCs4Q4dUDCBA&uact=5&oq=cristiano+ronaldo&gs_lp=Egxnd3Mtd2l6LXNlcnAiEWNyaXN0aWFubyByb25hbGRvMgUQABiABDIKEAAYgAQYigUYQzIFEAAYgAQyBRAAGIAEMgUQABiABDIFEAAYgAQyBRAAGIAEMgUQABiABDIFEAAYgAQyBRAAGIAESOAQUABYmxBwAXgAkAEAmAHeAaABuBWqAQYwLjE0LjK4AQPIAQD4AQGYAhGgAsYWwgIEECMYJ8ICCRAAGIAEGBgYCpgDAJIHBjEuMTMuM6AHwVM&sclient=gws-wiz-serp#vhid=-7UHW1Lud3bYdM&vssid=mosaic'

        # Last week and >800x800 images
        #return 'https://www.bing.com/images/search?q=' + query + '&qft=+filterui:age-lt10080+filterui:imagesize-custom_800_800&form=IRFLTR&first=1'

        return 'https://www.bing.com/images/search?q=' + query + '&go=Buscar&qs=ds&form=QBILPG&first=1'

        return 'https://www.bing.com/images/search?q=' + urllib.parse.quote_plus(query) + '&form=HDRSC3&first=1'

    """
    def save_image(self, link, file_path):
        request = urllib.request.Request(link, None, self.headers)
        image = urllib.request.urlopen(request, timeout = self.timeout).read()
        if not imghdr.what(None, image):
            print('[Error]Invalid image, not saving {}\n'.format(link))
            raise ValueError('Invalid image, not saving {}\n'.format(link))
        with open(str(file_path), 'wb') as f:
            f.write(image)
    """
    
    def download_image(self, link, output_filename):
        """
        This methods downloads the provided 'link' image and stores locally as
        'output_filename'. It will return the final filename with which it has
        been downloaded (as it can change due to wrong extension).
        """
        self.download_count += 1
        # Get the image link
        try:
            path = urllib.parse.urlsplit(link).path
            filename = posixpath.basename(path).split('?')[0]
            file_type = filename.split('.')[-1]
            if file_type.lower() not in IMAGE_FILE_EXTENSIONS:
                file_type = 'jpg'
                
            # We fix the extension if it is wrong
            if '.' in output_filename:
                extension = output_filename.split('.')[1]
                if extension != file_type:
                    output_filename.replace(extension, '.' + file_type)
            else:
                output_filename += '.' + file_type

            download_by_url(link, output_filename)
            # We append this link to avoid downloading it again
            self.downloaded_urls.add(link)

        except Exception as e:
            self.download_count -= 1
            print("[!] Issue getting: {}\n[!] Error:: {}".format(link, e))

        return output_filename

    def find_chrome(self, keywords, number_of_images):
        """
        image_scraper = GoogleImageScraper('', 'wip/', keywords, number_of_images)
        image_urls = image_scraper.find_image_urls()
        image_scraper.save_images(image_urls)

        return
        """
        # TODO: This is not working
    

        driver = start_chrome(True)
        #url = url = self.__get_url(keywords, 100)
        url = 'https://www.google.com/search?sca_esv=39bd196d7b057ff4&hl=es&sxsrf=ACQVn0-8K1iFPGSxADM3D9TRk5lpkoEV5A:1711109110553&q=' + keywords + '&uds=AMwkrPufgvfzFeIw5yilA4Gzc5SO9lVaLhgRHmmyPgNVqiyWpt_hRF_KcRkqwX4V1Ixye4HvydlY1Cv-RxhicPyEkaYN9AhDv9yeDfVCM1kWTfhEYzCUH1n5KMkCtCr83dibHEyrsdYFB09rfewbV-2PsHlpAxQkUqcQqO0Nswl38TiDcmJN_z4J94SD8eKY0Yx6_Aa2qTAWyLHuYV-xlZhimJRxB_QS2iSu6RC2hCa8GCd9HzMh8cubNzVF8XaauXGMqfjxBM-bVMAcgsvIHv6z7QwKxUFe6WRZkSTWMfpK8jlWaFprdh9pcL-iiNebRsbohdq5ESfr&udm=2&prmd=invsbmtz&sa=X&ved=2ahUKEwjJ7dKV6oeFAxX8XfEDHTpqAuIQtKgLegQIDRAB&biw=1368&bih=785&dpr=2'
        go_to_and_wait_loaded(driver, url)
        time.sleep(10)

        #href="https://www.google.com/imgres?imgurl=https%3A%2F%2Fupload.wikimedia.org%2Fwikipedia%2Fcommons%2Fe%2Fef%2FIbai_Llanos_in_2020.jpg&amp;tbnid=oA2nnqJVRGvgrM&amp;vet=12ahUKEwi73cLrgoiFAxVVov0HHRs-BdMQMygAegQIARBu..i&amp;imgrefurl=https%3A%2F%2Fen.wikipedia.org%2Fwiki%2FIbai_Llanos&amp;docid=7HS4A3h3iDalNM&amp;w=620&amp;h=680&amp;q=ibai&amp;hl=es&amp;ved=2ahUKEwi73cLrgoiFAxVVov0HHRs-BdMQMygAegQIARBu"

        # https://www.google.com/imgres?imgurl=https%3A%2F%2Fupload.wikimedia.org%2Fwikipedia%2Fcommons%2Fe%2Fef%2FIbai_Llanos_in_2020.jpg&tbnid=oA2nnqJVRGvgrM&vet=12ahUKEwje6vS-gYiFAxXtnf0HHar3C7YQMygAegQIARBv..i&imgrefurl=https%3A%2F%2Fen.wikipedia.org%2Fwiki%2FIbai_Llanos&docid=7HS4A3h3iDalNM&w=620&h=680&q=ibai&hl=es&ved=2ahUKEwje6vS-gYiFAxXtnf0HHar3C7YQMygAegQIARBv

        #elements = driver.find_elements(By.XPATH, "//a[contains(@href,'/imgres?q=')]")
        #elements = driver.find_elements(By.XPATH, "//a[contains(@href, 'https://www.google.com/imgres?')]")
        #elements = driver.find_elements(By.XPATH, "//h3[contains(@class, 'ob5Hkd')]")#/*[1]
        #elements = driver.find_elements(By.CLASS_NAME, 'ob5Hkd')
        #elements = driver.find_elements(By.TAG_NAME, 'a')
        #elements = driver.find_elements(By.XPATH, "//a[contains(@data-navigation, 'server')]")#/*[1]
        #elements = driver.find_elements(By.XPATH, '//div[contains(@role, "list-item")]')
        # press tab 40 times and get focused
        actions = ActionChains(driver) 
        actions.send_keys(Keys.TAB * 50)
        actions.perform()
        elem = driver.switch_to.active_element
        print(elem.tag_name)
        elem.click()
        time.sleep(2)
        eless = driver.find_element(By.XPATH, '//a[contains(@aria-hidden, "false")]')
        print(len(eless))
        print(eless[0].get_attribute('href'))
        if elem.tag_name == 'a':
            print(elem.get_attribute('href'))
            print(elem.text)
        time.sleep(60)
        return []
        for element in elements:
            href = element.get_attribute('href')
            if href:
                print(href)
                if '/imgres' in href:
                    print(href)
            else:
                print(element.text)
        return []
        urls = []
        for element in elements:
            # Url is like this one:
            # https://www.google.com/imgres?q=cristiano%20ronaldo&imgurl=https%3A%2F%2Fupload.wikimedia.org%2Fwikipedia%2Fcommons%2Fd%2Fd7%2FCristiano_Ronaldo_playing_for_Al_Nassr_FC_against_Persepolis%252C_September_2023_%2528cropped%2529.jpg&imgrefurl=https%3A%2F%2Fes.wikipedia.org%2Fwiki%2FCristiano_Ronaldo&docid=ty38cmTK8QLflM&tbnid=AB35wZltZBgFqM&vet=12ahUKEwjc_uGW6oeFAxVcQ_EDHXxbBF4QM3oECBYQAA..i&w=805&h=1053&hcb=2&itg=1&ved=2ahUKEwjc_uGW6oeFAxVcQ_EDHXxbBF4QM3oECBYQAA
            url = element.get_attribute('src')
            url = url.split('&')[1] # imgurl=...
            url = url.split('=')[1]
            url = urllib.parse.unquote(url)
            urls.append(url)

        return urls

    def find(self, keywords, number_of_images):
        """
        This will look for 'number_of_images' images using the Bing browser and will
        return their links. This method will skip the previously downloaded images
        to avoid duplicities, if that options is active.
        """
        driver = start_chrome()
        url = url = self.__get_url(keywords, 100)
        go_to_and_wait_loaded(driver, url)
        buttons = driver.find_elements(By.XPATH, "//img[contains(@class, 'mimg')]")
        links = []
        for button in buttons:
            links.append(button.get_attribute('src'))

        if self.ignore_repeated:
            links = list(filterfalse(self.downloaded_urls.__contains__, links))

        # Force limit here
        links = links[:number_of_images]

        return links

        # I force the 'limit' to be 100 and then I manually remove
        url = self.__get_url(keywords, 100)
        request = urllib.request.Request(url, None, headers = self.headers)
        response = urllib.request.urlopen(request)
        html = response.read().decode('utf8')
        if html ==  "":
            return None
        
        # TODO: Implement a way of requesting in other pages
        # This below works with the first option in '__get_url()'
        #links = re.findall('murl&quot;:&quot;(.*?)&quot;', html)
        links = re.findall('src2=&quot;(.*?)&quot;', html)

        if self.ignore_repeated:
            links = list(filterfalse(self.downloaded_urls.__contains__, links))

        # Force limit here
        links = links[:number_of_images]

        return links
    
    def find_and_download(self, keywords, number_of_images):
        """
        This will look for 'number_of_images' images using the Bing browser and will
        download those images. This method will skip the previously downloaded images
        to avoid duplicities, if that options is active.
        """
        #urls = self.find(keywords, number_of_images)
        urls = self.find(keywords, number_of_images)
        for url in urls:
            self.download_image(url)
    
    """
    def run(self):
        while self.download_count < self.limit:
            # Parse the page source and download pics
            request_url = 'https://www.bing.com/images/async?q=' + urllib.parse.quote_plus(self.query) \
                          + '&first=' + str(self.page_counter) + '&count=' + str(self.limit) \
                          + '&adlt=' + self.adult + '&qft=' + ('' if self.filter is None else self.__get_filter(self.filter))
            request = urllib.request.Request(request_url, None, headers=self.headers)
            response = urllib.request.urlopen(request)
            html = response.read().decode('utf8')
            if html ==  "":
                print("[%] No more images are available")
                break
            links = re.findall('murl&quot;:&quot;(.*?)&quot;', html)

            for link in links:
                if self.download_count < self.limit and link not in self.seen:
                    self.seen.add(link)
                    self.download_image(link)

            self.page_counter += 1
    """


from yta_multimedia.video.generation.rrss.discord.classes.discord_conversation_message import DiscordConversationMessage
from moviepy.editor import concatenate_videoclips
from typing import Union

class DiscordConversationVideoGenerator:
    def __init__(self):
        pass

    def generate_clip(self, messages: list[DiscordConversationMessage], output_filename: Union[str, None] = None):
        """
        Generates a video clip with the provided messages. This method
        will return the clip and also write it locally if 'output_filename'
        is provided.
        """
        if not messages:
            raise Exception('No "messages" provided.')
        
        clips = []
        for index, message in enumerate(messages):
            # If previous clip is from the same author, concat the 'text'
            if len(clips) > 0:
                previous_message = messages[index - 1]
                if previous_message.character == message.character:
                    message.text = previous_message.text + '\\n' + message.text

            clips.append(message.generate_clip())

        clip = concatenate_videoclips(clips).set_fps(60)

        if output_filename:
            clip.write_videofile(output_filename)

        return clip


from openvoice import se_extractor
from openvoice.api import ToneColorConverter
from melo.api import TTS
from pathlib import Path
from dotenv import load_dotenv
from yta_general_utils.file_processor import get_project_abspath

import os
import torch

PROJECT_ABSOLUTE_PATH = get_project_abspath

def narrate(text, speed = 1.0, output_filename = None):
    """
    Narrates the provided 'text' at the provided 'speed' with the MeloTTS
    library. The file will be saved as 'output_filename'.

    # TODO: @definitive_cantidate
    """
    if not output_filename:
        return None
    
    # TODO: Check if speed is valid
    
    # This below will automatically choose GPU if available
    device = 'auto' 
    model = TTS(language = 'ES', device = device)
    speaker_ids = model.hps.data.spk2id
    model.tts_to_file(text, speaker_ids['ES'], output_filename, speed = speed)

def clone_voice(input_filename):
    CHECKPOINTS_PATH = (Path(__file__).parent.parent.__str__() + '/resources/openvoice/checkpoints_v2/').replace('\\', '/')
    
    ckpt_converter = CHECKPOINTS_PATH + 'converter'
    device = 'cuda:0' if torch.cuda.is_available() else 'cpu'
    tone_color_converter = ToneColorConverter(f'{ckpt_converter}/config.json', device = device)
    tone_color_converter.load_ckpt(f'{ckpt_converter}/checkpoint.pth')
    source_se = torch.load(f'{CHECKPOINTS_PATH}/base_speakers/ses/es.pth', map_location = device)
    # This will generate a 'se.pth' file and some wavs that are the cloned voice
    target_se, audio_name = se_extractor.get_se(input_filename, tone_color_converter, vad = False)


def imitate_voice(text, input_filename = None, output_filename = None):
    """
    This method imitates the 'input_filename' provided voice and
    generates a new narration of the provided 'text' and stores it
    as 'output_filename'.

    The provided 'input_filename' must be a valid audio file that
    contains a clear narration to be imitated.

    # TODO: @definitive_cantidate
    """
    if not input_filename:
        return None
    
    if not output_filename:
        return None
    
    CHECKPOINTS_PATH = (Path(__file__).parent.parent.__str__() + '/resources/openvoice/checkpoints_v2/').replace('\\', '/')
    
    ckpt_converter = CHECKPOINTS_PATH + 'converter'
    device = 'cuda:0' if torch.cuda.is_available() else 'cpu'
    tone_color_converter = ToneColorConverter(f'{ckpt_converter}/config.json', device = device)
    tone_color_converter.load_ckpt(f'{ckpt_converter}/checkpoint.pth')

    source_se = torch.load(f'{CHECKPOINTS_PATH}/base_speakers/ses/es.pth', map_location = device)
    target_se, audio_name = se_extractor.get_se(input_filename, tone_color_converter, vad = False)

    # This below is for testing
    # audio_segs is the number of audio segments created
    # se_save_path is the path in which se.pth file has been saved
    # TODO: Need to know the path in which everything is saved to detect audio
    # segments number and also to be able to load the 'se.pth' file
    path = PROJECT_ABSOLUTE_PATH + 'processed/narracion_irene_albacete_recortado_v2_OMR2KXcN3jYVFUsb'
    tone_color_converter.extract_se(30, se_save_path = path), 'narracion_irene_albacete_recortado_v2_OMR2KXcN3jYVFUsb'
    # TODO: Check what is 'target_se' to check if it is a string and we can
    # point the 'se.pth' file, because I don't already understand how it works
    # This above is for testing

    # We generate a narration to obtain it but with the 'input_filename' voice
    source_filename = 'tmp.wav'
    narrate(text, output_filename = source_filename)

    encode_message = "@MyShell"
    tone_color_converter.convert(
        audio_src_path = source_filename, 
        src_se = source_se, 
        tgt_se = target_se, 
        output_path = output_filename,
        message = encode_message)
    
    # TODO: Remove tmp file 'source_filename'
    try:
        os.remove('tmp.wav')
    except:
        pass



def __test():
    # TODO: This must be deleted, I keep it to ensure nothing will fail in the future
    # TODO: Took from here (https://github.com/myshell-ai/OpenVoice/blob/main/demo_part3.ipynb)
    PATH = 'C:/Users/dania/Desktop/PROYECTOS/yta-ai-utils/yta_ai_utils/'

    ckpt_converter = PATH + 'resources/openvoice/checkpoints_v2/converter'
    device = "cuda:0" if torch.cuda.is_available() else "cpu"
    output_dir = 'output/openvoice'

    tone_color_converter = ToneColorConverter(f'{ckpt_converter}/config.json', device = device)
    tone_color_converter.load_ckpt(f'{ckpt_converter}/checkpoint.pth')

    os.makedirs(output_dir, exist_ok = True)

    reference_speaker = PATH + 'resources/test.m4a' # This is the voice you want to clone
    target_se, audio_name = se_extractor.get_se(reference_speaker, tone_color_converter, vad = False)

    texts = {
        'EN_NEWEST': "Did you ever hear a folk tale about a giant turtle?",  # The newest English base speaker model
        'EN': "Did you ever hear a folk tale about a giant turtle?",
        'ES': "El resplandor del sol acaricia las olas, pintando el cielo con una paleta deslumbrante.",
        'FR': "La lueur dorée du soleil caresse les vagues, peignant le ciel d'une palette éblouissante.",
        'ZH': "在这次vacation中，我们计划去Paris欣赏埃菲尔铁塔和卢浮宫的美景。",
        'JP': "彼は毎朝ジョギングをして体を健康に保っています。",
        'KR': "안녕하세요! 오늘은 날씨가 정말 좋네요.",
    }

    src_path = f'{output_dir}/tmp.wav'

    # Basic (no cloning) below
    speed = 1.0

    for language, text in texts.items():
        model = TTS(language=language, device=device)
        speaker_ids = model.hps.data.spk2id
        
        for speaker_key in speaker_ids.keys():
            speaker_id = speaker_ids[speaker_key]
            speaker_key = speaker_key.lower().replace('_', '-')
            
            source_se = torch.load(f'{PATH}checkpoints_v2/base_speakers/ses/{speaker_key}.pth', map_location=device)
            model.tts_to_file(text, speaker_id, src_path, speed = speed)
            save_path = f'{output_dir}/output_v2_{speaker_key}.wav'

            # Run the tone color converter
            encode_message = "@MyShell"
            tone_color_converter.convert(
                audio_src_path=src_path, 
                src_se=source_se, 
                tgt_se=target_se, 
                output_path=save_path,
                message=encode_message)


"""
Mobject to represent a video inside of Manim. This video
Mobject is a modified ImageMobject that updates the 
content on each iteration.

This mobject is working properly with 'cairo' render and
using the 'ImageMobject' class for the inheritance.

TODO: Make this work with 'opengl' render by replacing 
the 'ImageMobject' with 'OpenGLImageMobject', but this
is returning a 'astype' exception when formating the
content with PIL lib in '.astype('uint8')'.

Thanks: https://gist.github.com/uwezi/faec101ed5d7c20222b33eee4b6c7d63
"""
from manim.mobject.opengl.opengl_image_mobject import OpenGLImageMobject
from manim import *
from dataclasses import dataclass
from PIL import Image


import cv2

@dataclass
class VideoStatus:
    time: float = 0
    videoObject: cv2.VideoCapture = None

    def __deepcopy__(self, memo):
        return self

# TODO: Try to use 'OpenGLImageMobject' to be able
# to render in a 3D envinroment
# If trying with 'OpenGLImageMobject' you should
# use the 'opengl' animator , but this is failing
class VideoOpenGLMobject(OpenGLImageMobject):
    '''
    Mobject that represents a video as a sequence of 
    OpenGLImageMobjects that are updated as a video: frame by frame
    (it is not working).

    This VideoMobject has been created to work with 'Opengl' render.
    The positioning is perfect in 2D and 3D scenes, but the video
    is not being displayed. I cannot update the image that has to
    be shown in each iteration. It seems to be detecting the image
    in the updater, but it is not being updated by their parent
    classes.

    Parameters
    ----------
    filename
        the filename of the video file

    loop
        (optional) replay the video from the start in an endless loop

    https://discord.com/channels/581738731934056449/1126245755607339250/1126245755607339250
    2023-07-06 Uwe Zimmermann & Abulafia
    2024-03-09 Uwe Zimmermann
    '''
    def __init__(self, filename = None, loop = True, **kwargs):
        self.filename = filename
        self.speed = 1
        self.loop = loop
        self._id = id(self)
        self.status = VideoStatus()
        self.status.videoObject = cv2.VideoCapture(filename)

        self.status.videoObject.set(cv2.CAP_PROP_POS_FRAMES, 1)
        ret, frame = self.status.videoObject.read()
        
        if ret:
            frame = cv2.cvtColor(frame, cv2.COLOR_BGR2RGB)            
            img = Image.fromarray(frame)
        else:
            img = Image.fromarray(np.uint8([[63, 0, 0, 0],
                                        [0, 127, 0, 0],
                                        [0, 0, 191, 0],
                                        [0, 0, 0, 255]
                                        ]))
        self.size = img.size

        # Image to np.darray to avoid fail in 'OpenGLImageMobject'
        # that raise an Exception in this line:
        # Image.fromarray(image_file.astype("uint8"))
        img = np.array(img)

        # TODO: We need to manually fix the 'OpenGLImageMobject' and 
        # set the '.resize(...)' method to *2 instead of *200 or I
        # receive a MemoryError
        super().__init__(img, **kwargs)

        if ret:
            self.add_updater(self.video_updater)

    def video_updater(self, mobj, dt):
        """
        This method updates the image (frame) that needs to be
        shown in the animation attending to the moment.
        """
        if dt == 0:
            return
        
        status = self.status
        status.time += 1000 * dt * mobj.speed
        self.status.videoObject.set(cv2.CAP_PROP_POS_MSEC, status.time)
        ret, frame = self.status.videoObject.read()

        # Go to frame 0 if ended and loop requested
        if (ret == False) and self.loop:
            status.time = 0
            self.status.videoObject.set(cv2.CAP_PROP_POS_MSEC, status.time)
            ret, frame = self.status.videoObject.read()

        if ret:
            frame = cv2.cvtColor(frame, cv2.COLOR_BGR2RGB)  # needed here?        
            img = Image.fromarray(frame)

            # Update the content (the image shown)
            # TODO: This is not working, the parents don't
            # do nothing to actually upgrade the current
            # image and I don't know what else I can try
            mobj.image = img
            mobj.image_file = img
           


from yta_general_utils.tmp_processor import create_tmp_filename
from yta_general_utils.file_downloader import download_image
from PIL import Image, ImageFont, ImageDraw, ImageOps
from random import randrange as random_randrange

import textwrap
import numpy as np

def create_tripadvisor_review_image(review: str = None, avatar_url: str = None, username: str = None, city: str = None, contributions_number: int = random_randrange(30), rating_stars: int = (random_randrange(5) + 1), title: str = None, output_filename: str = None):
    # TODO: This must be set by the user in .env I think
    __FONTS_PATH = 'C:/USERS/DANIA/APPDATA/LOCAL/MICROSOFT/WINDOWS/FONTS/'

    # TODO: Do more checkings please
    # TODO: Check contributions_number is in between 1 and 30 and is a number
    # TODO: Check rating_stars is a number between 1 and 5
    # TODO: Check title exists
    # TODO: Check review exists

    bold_font = ImageFont.truetype(__FONTS_PATH + 'HUMANIST_777_BOLD_BT.TTF', 22, encoding = 'unic')
    normal_font = ImageFont.truetype(__FONTS_PATH + 'HUMANIST-777-FONT.OTF', 16, encoding = 'unic')
    review_font = ImageFont.truetype(__FONTS_PATH + 'HUMANIST-777-FONT.OTF', 20, encoding = 'unic')

    # Create white review background (440 is h, 800 is w)
    img = np.zeros((440, 800, 3), dtype = np.uint8)
    img.fill(255)
    img = Image.fromarray(img).convert('RGBA')
    draw = ImageDraw.Draw(img)

    # Place avatar image circle at (24, 31)
    image_filename = create_tmp_filename('tmp_avatar.png')
    if not avatar_url:
        avatar_url = 'https://avatar.iran.liara.run/public'
    download_image(avatar_url, image_filename)
    avatar_image = Image.open(image_filename)

    # Make image fit and then mask with circle
    avatar_image = ImageOps.fit(avatar_image, (64, 64))
    mask = Image.new('L', avatar_image.size, 0)
    circle_draw = ImageDraw.Draw(mask)
    circle_draw.ellipse((0, 0, 64, 64), fill = 255)
    result = Image.new('RGBA', avatar_image.size, (255, 255, 255))
    result.paste(avatar_image, (0, 0), mask)

    img.paste(result, (24, 31))

    # I check if I need some fake information
    fake_data = None
    if not username or not city:
        # TODO: Move this below to a 'get_json_from_url' or similar
        # TODO: Create the new 'yta-data-provider' library and include this
        import requests
        response = requests.get('https://fakerapi.it/api/v1/persons?_locale=es_ES')
        data = response.json()
        fake_data = data['data'][0] # First person generated data

    # Put reviewer name at (103, 33)
    if not username:
        username = fake_data['firstname'] + ' ' + fake_data['lastname']
    draw.text((103, 33), username, font = bold_font, fill = (84, 84, 84), line_height = '19px', width = 700)
    
    # Write city and contributions at (103, 65)
    if not city:
        city = fake_data['address']['city']
    # TODO: Check contributions_number is a number
    draw.text((103, 65), city + ' · ' + str(contributions_number) + ' contribuciones', font = normal_font, fill = (84, 84, 84), line_height = '19px', width = 700)

    # Build the evaluation (in green circles) at (24, 126)
    def __draw_filled_ellipse(xy):
        return draw.ellipse(xy, fill = (0, 170, 108), outline = (0, 170, 108))
    def __draw_unfilled_ellipse(xy):
        return draw.ellipse(xy, fill = (255, 255, 255), outline = (0, 170, 108))
    
    x = 24
    for i in range(5):
        if i < rating_stars:
            __draw_filled_ellipse((x, 126, x + 24, 126 + 24))
        else:
            __draw_unfilled_ellipse((x, 126, x + 24, 126 + 24))
        x += 24 + 6

    # Place the review title at (24, 164)
    if not title:
        # TODO: Fake review title with AI
        title = 'Menuda pasada'
    draw.text((24, 164), title, font = bold_font, fill = (84, 84, 84), line_height = '19px')

    # Place the review date at (24, 199)
    # TODO: Implement 'date' parameter
    # TODO: Implement 'visit_type' options
    draw.text((24, 199), "jul 2019 · En pareja", font = normal_font, fill = (84, 84, 84), line_height = '19px')

    # Place the whole wrapped text starting at (24, 440 - 180)
    # See this: (https://gist.github.com/turicas/1455973) to wrap text
    if not review:
        # TODO: Fake review with AI
        review = 'Estuvimos todo el viaje pensando en que ibamos a poder visitar por fin este sitio, y la verdad es que no ha defraudado nada. Lo disfrutamos de principio a fin y la atención fue espectacular. Llevábamos unas espectativas bastante altas pero, sin lugar a dudas, las han superado. Encantados!!'
    lines = textwrap.wrap(review, width = 76)
    y_text = 440 - 180
    for line in lines:
        left, top, right, bottom = review_font.getbbox(line)
        height = bottom - top
        draw.text((24, y_text), line, font = review_font, fill = (84, 84, 84))
        y_text += height + 8

    # Place the number of likes: 700, 50
    #draw.text((700, 50), "3", font = normal_font, fill = (84, 84, 84), line_height = '19px')

    if not output_filename:
        output_filename = create_tmp_filename('tmp_tripadvisor_review.png')
    img.save(output_filename, 'png')

    # TODO: Why not returning the image instead (?)
    return output_filename


from yta_multimedia.video.generation.manim.utils.dimensions import height_to_manim_height

class LoadingBarImage:
    def __init__(self, image_filename: str, y: int, start_percentage: int, end_percentage: int):
        """
        An object that represents an image that will be used in the loading
        bar to improve the animation quality. This image will be static or
        dynamic. If 'start_percentage' and 'end_percentage' are the same,
        the image will be static at that percentage point. If not, it will
        move from 'start_percentage' to 'end_percentage'.

        The 'y' parameter is the vertical distance from the loading bar. 
        This is useful to position the image above, below or inside the
        bar.
        """
        if not image_filename: # TODO: Check if valid Image
            raise Exception('"image_filename" parameter must be a valid image filename string')
        
        if start_percentage == None or start_percentage < 0 or start_percentage > 100:
            raise Exception('"start_percentage" parameter must be a valid int between [0, 100]')
        
        if end_percentage == None or end_percentage < 0 or end_percentage > 100:
            raise Exception('"start_percentage" parameter must be a valid int between [0, 100]')
        
        if y == None or y < -100 or y > 100:
            raise Exception('"y" parameter must be a valid int between [-100, 100]')

        self.image_filename = image_filename
        self.start_percentage = start_percentage
        self.end_percentage = end_percentage
        self.y = height_to_manim_height(y)
        # This is for the right movement
        self.start_x = 0
        self.end_x = 0

    def toJSON(self):
        return {
            'image_filename': self.image_filename,
            'y': self.y,
            'start_percentage': self.start_percentage,
            'end_percentage': self.end_percentage
        }


from moviepy.editor import VideoFileClip
from yta_general_utils.type_checker import variable_is_type
from yta_general_utils.file_processor import file_is_video_file
from yta_general_utils.tmp_processor import create_tmp_filename
from typing import Union

def crop_video_using_key_frame_second(video_input: Union[VideoFileClip, str], key_frame_second: float, duration: float, output_filename: Union[str, None] = None):
    if not video_input:
        return None
    
    if variable_is_type(video_input, str):
        if not file_is_video_file(video_input):
            return None
        
        video_input = VideoFileClip(video_input)

    if not key_frame_second or key_frame_second < 0 or key_frame_second > video_input.duration:
        # We use the middle of the video as the key frame
        key_frame_second = video_input.duration / 2
    if not duration:
        duration = video_input.duration
    if duration > video_input.duration or duration <= 0:
        duration = video_input.duration

    if duration < video_input.duration:
        start_second = 0
        end_second = video_input.duration
        # Only if we have to crop it already
        half_duration = duration / 2
        if key_frame_second - half_duration < 0:
            # Start in 0.0
            start_second = 0
            end_second = duration
        elif key_frame_second + half_duration > video_input.duration:
            # End in 'video_input.duration'
            start_second = video_input.duration - duration
            end_second = video_input.duration
        else:
            # Use 'key_frame_second' as center
            start_second = key_frame_second - half_duration
            end_second = key_frame_second + half_duration

        video_input = video_input.subclip(start_second, end_second)

    if output_filename:
        # TODO: Check extension
        tmp_audiofilename = create_tmp_filename('temp-audio.m4a')
        # TODO: Do I really need all those parameters?
        video_input.to_videofile(
            output_filename,
            codec = "libx264",
            temp_audiofile = tmp_audiofilename,
            remove_temp = True,
            audio_codec = 'aac' # pcm_s16le or pcm_s32le
        )

    return video_input


"""
1. You need to install Ollama to handle llama models. Just download
and install from this website: https://ollama.com

# This one (https://www.youtube.com/watch?v=EBUMxu2hl34) says that
you can use openai.chat to communicate to the ollama model

2. You must install 'ollama' python package to communicate with the
models you download and use

# Based on this (https://www.youtube.com/watch?v=4Jpltb9crPM) I am
not able to execute it because it never ends loading llava I think

https://llava-vl.github.io/
"""
import ollama

def describe_image_NOT_WORKING(image_filename: str):
    """
    This method will return a description of the provided 
    'image_filename' based on the ollama 'llava' model.

    TODO: This is not working because of my pc limitations.
    It cannot load the resources due to memory capacity.
    """
    res = ollama.chat(
        model = 'llava',
        messages = [
            {
                'role': 'user',
                'content': 'Describe this image',
                'images': [
                    image_filename
                ]
            }
        ]
    )

    response_content = res['message']['content']

    return response_content

# print('here trying to')
# path = 'C:/Users/dania/Desktop/suscribirte_imagen_guapa.png'
# print(describe_image(path))
# print('ended?')


from moviepy.editor import AudioFileClip

def append_audio_clip_to_audio_clip(base_audio_clip: AudioFileClip, new_audio_clip: AudioFileClip, start_time: float = 0.0, output_filename: str = None):
    """
    We receive a 'base_audio_clip' in which we will append the 'new_audio_clip'
    starting at 'start_time' seconds since the first one started. A combined 
    audio will be generated and returned.

    If 'output_filename' is provided, it will be also written in that 
    destination.
    """
    if start_time < 0:
        start_time = 0

    from moviepy.editor import CompositeAudioClip

    combined = CompositeAudioClip([base_audio_clip, new_audio_clip.set_start(start_time)])
    # TODO: Why 441000 here below?
    combined.fps = 44100

    if output_filename:
        combined.write_audiofile(output_filename)

    return combined

def append_audio_file_to_audio_file(base_audio_filename: str, new_audio_filename: str, start_time: float = 1.0, output_filename: str = None):
    """
    We receive an audio file called 'base_audio_filename' and we append the provided
    'new_audio_filename' to that existing audio to sound over it. That second 
    sound will start in the 'start_time' second of the first 'audio_filename'.

    If 'output_filename' provided, it will write the new combined audio. If not,
    it will only return the new CompositeAudioClip object.

    # TODO: Is this method really necessary (?) I think it is not...
    """
    from yta_general_utils.file_processor import file_is_audio_file

    if not base_audio_filename:
        return None
    
    if not file_is_audio_file(base_audio_filename):
        return None
    
    if not new_audio_filename:
        return None
    
    if not file_is_audio_file(new_audio_filename):
        return None
    
    if start_time < 0:
        return None

    return append_audio_clip_to_audio_clip(AudioFileClip(base_audio_filename), AudioFileClip(new_audio_filename), start_time, output_filename)


# TODO: Remove this when new greenscreen system has been
# completely implemented and this file doesn't make sense
from greenscreen.titled_green_screen_image import TitledGreenScreenImage

class GreenScreenImage(TitledGreenScreenImage):
    """
    Green screen of custom size that is placed over a background image of 1920x1080.
    """
    def __init__(self, width = 1344, height = 756):
        super().__init__('', width, height)



from moviepy.editor import VideoFileClip
from yta_general_utils.type_checker import variable_is_type
from yta_general_utils.file_processor import file_is_video_file
from typing import Union

def get_all_frames_from_video(video: Union[VideoFileClip, str], output_folder: Union[str, None] = None):
    """
    This method will get all the frames of the provided 'video' and
    will return them. It 'output_folder' is provided, they will be
    also stored locally in that folder with the 'frameXXXXX.png' 
    name, from 0 to the last frame that exist.

    @param
        **video**
        The video from which we want to get all the frames
    """
    if not video:
        return None
    
    if output_folder and not output_folder.endswith('/'):
        output_folder += '/'
    
    if variable_is_type(video, str):
        if not file_is_video_file(video):
            return None
        
        video = VideoFileClip(video)

    # TODO: Maybe enhance this by first getting them and writing
    # manually with other method, and then returning
    if output_folder:
        video.write_images_sequence(output_folder + 'frame%05d.png')

    frame_numbers = [i for i in range((int) (video.fps * video.duration))]

    return get_frames_from_video_by_frame_numbers(video, frame_numbers)

# TODO: This method should be in another place as its behaviour is
# very different as the single one below. This is for summarizing
# a video or to help other functionalities like that
def get_frames_from_video_by_frame_numbers(video: Union[VideoFileClip, str], frame_numbers = [], output_folder: str = None):
    """
    This method will obtain the frame corresponding to the provided
    'frame_numbers' of the also provided 'video'. They will be 
    returned and if 'output_folder' is given, they will be stored
    locally.

    @param
        **frame_numbers**
        A list containing the frame numbers we want to get, that 
        need to be available in the provided 'video'.
    """
    if not video:
        return None
    
    # TODO: Check 'frame_numbers' better
    if len(frame_numbers) == 0:
        return None
    
    if output_folder and not output_folder.endswith('/'):
        output_folder += '/'
    
    if variable_is_type(video, str):
        if not file_is_video_file(video):
            return None
        
        video = VideoFileClip(video)

    # Obtain the max frame number and check if it is available
    if max(frame_numbers) > ((int) (video.fps * video.duration)):
        raise Exception('Some frame requested does not exist in the provided video.')
    
    frames = []
    for frame_number in frame_numbers:
        output_filename = None
        if output_folder:
            output_filename = 'frame' + str(frame_number).zfill(5) + '.png'
        frames.append(get_frame_from_video_by_frame_number(video, frame_number, output_filename))
        
    return frames

def get_frame_from_video_by_frame_number(video: Union[VideoFileClip, str], frame_number: int, output_filename: Union[str, None] = None):
    """
    Extracts the frame 'frame_number' from the provided 'video' that will
    be stored locally as 'output_filename' if that parameter is provided
    or will not if it is None. This method will return the frame as

    @param
        **video**
        The video from which you want to get the frame.

    @param
        **frame_number**
        The frame that you want to extract from the video. This must be
        an int. If you want to get the first frame, this parameter must
        be 0. If you want to get the 26th frame, pass 26 as parameter.

    @param
        **output_filename**
        The name to store the frame locally. None value will make being
        not stored, only returned.
    """
    if not video:
        return None
    
    if frame_number < 0:
        return None
    
    if variable_is_type(video, str):
        if not file_is_video_file(video):
            return None
        
        video = VideoFileClip(video)

    if frame_number > (int) (video.fps * video.duration):
        return None

    return get_frame_from_video_by_time(video, frame_number * 1.0 / video.fps, output_filename)

def get_frame_from_video_by_time(video: Union[VideoFileClip, str], time: float, output_filename: str = None):
    """
    Extracts a frame from the provided 'video' that is exactly at the
    also provided 'time' moment. The frame will be stored as 
    'output_filename'.

    @param
        **time**
        The moment of the clip in which the frame is shown. This should be
        something like "frame_number * 1.0 / self.video.fps" to work well.
        Passing 0 will return the first frame of the video, and passing 2
        will return the frame in the second 2 of the video.
    """
    if not video:
        return None
    
    if time < 0:
        return None
    
    if variable_is_type(video, str):
        if not file_is_video_file(video):
            return None
        
        video = VideoFileClip(video)

    if time > video.duration:
        return None
    
    if output_filename:
        video.save_frame(t = time, filename = output_filename)

    return video.get_frame(t = time)


from moviepy.editor import VideoFileClip, CompositeVideoClip, ImageClip, concatenate_videoclips
from yta_multimedia.video.edition.effect.moviepy.fade_in_moviepy_effect import FadeInMoviepyEffect
from yta_multimedia.video.edition.effect.moviepy.fade_out_moviepy_effect import FadeOutMoviepyEffect
from typing import Union

class BlinkMoviepyEffect:
    """
    This method gets the first frame of the provided 'clip' and returns a
    new clip that is an incredible 'sad_moment' effect with black and white
    filter, zoom in and rotating effect and also sad violin music.

    The 'duration' parameter is to set the returned clip duration, but the
    default value is a perfect one.

    TODO: Maybe adjust the duration as I do with 'PhotoMoviepyEffect' in 
    which I set a default effect duration and I put the effect in the 
    middle of the provided clip.
    """
    __parameters = {}

    def __init__(self, clip: Union[VideoFileClip, CompositeVideoClip, ImageClip], duration: float = 1, color = [0, 0, 0]):
        self.__clip = clip
        self.__parameters['duration'] = duration
        self.__parameters['final_color'] = color

    def __process_parameters(self):
        if not self.__parameters['duration']:
            self.__parameters['duration'] = 1
        else:
            # Zoom is by now limited to [0.1 - 5] ratio
            if self.__parameters['duration'] > 5:
                self.__parameters['duration'] = 5
            elif self.__parameters['duration'] <= 0.1:
                self.__parameters['duration'] = 0.1

        if not self.__parameters['final_color']:
            self.__parameters['final_color'] = [0, 0, 0]
        # TODO: Check color is a valid value.

        return self.__parameters
    
    def apply(self):
        """
        Applies the effect to the provided 'clip' and with the also
        provided parameters needed by this effect.
        """
        if not self.__clip:
            return None
        
        self.__process_parameters()
        
        half_duration = self.__clip.duration / 2
        self.__clip = concatenate_videoclips([
            FadeOutMoviepyEffect(self.__clip.subclip(0, half_duration), duration = half_duration, color = self.__parameters['final_color']).apply(),
            FadeInMoviepyEffect(self.__clip.subclip(half_duration, self.__clip.duration), duration = half_duration, color = self.__parameters['final_color']).apply(),
        ])

        return self.__clip


from yta_multimedia.video.frames import get_frame_from_video_by_frame_number
from yta_general_utils.type_checker import variable_is_type
from yta_general_utils.file_processor import file_is_video_file
from moviepy.editor import VideoFileClip, ImageClip, CompositeVideoClip, concatenate_videoclips
from typing import Union

class StopMotionVideoEffect:
    """
    Creates a Stop Motion effect in the provided video by dropping the frames
    per second but maintaining the original ratio.
    """
    def __init__(self, video: Union[VideoFileClip, ImageClip, CompositeVideoClip, str]):
        if not video:
            raise Exception('No "video" provided.')
        
        if variable_is_type(video, str):
            if not file_is_video_file(video):
                raise Exception('Provided "video" is not a valid video file.')
            
            video = VideoFileClip(video)

        self.video = video

    def apply(self):
        # This value is this one by default by now
        FRAMES_TO_JUMP = 5

        clips = []
        for frame_number in range((int) (self.video.fps * self.video.duration)):
            if frame_number % FRAMES_TO_JUMP == 0:
                frame = get_frame_from_video_by_frame_number(self.video, frame_number)
                clips.append(ImageClip(frame, duration = FRAMES_TO_JUMP / self.video.fps).set_fps(self.video.fps))

        return concatenate_videoclips(clips).set_audio(self.video.audio).set_fps(self.video.fps)


    


from moviepy.editor import vfx, VideoFileClip, CompositeVideoClip, ImageClip
from typing import Union

class ChangeSpeedMoviepyEffect:
    """
    This effect changes the speed of the video. A 'speed_factor'
    of 2 means a 2x speed (0:10s -> 0:05s). A 'speed_factor' of
    0.5 means a 0.5x speed (0:10s -> 0:20s)
    """
    __MOVIEPY_EFFECT_NAME = 'speedx'
    __parameters = {}

    def __init__(self, clip: Union[VideoFileClip, CompositeVideoClip, ImageClip], speed_factor: float = None):
        self.__clip = clip
        self.__parameters['speed_factor'] = speed_factor

    def __get_moviepy_vfx_effect(self):
        return getattr(vfx, self.__MOVIEPY_EFFECT_NAME, None)
    
    def __process_parameters(self):
        if not self.__parameters['speed_factor']:
            self.__parameters['speed_factor'] = 1
        else:
            # This is limited to x10 and x0.1
            if self.__parameters['duration'] > 10:
                self.__parameters['duration'] = 1
            elif self.__parameters['duration'] <= 0.1:
                self.__parameters['duration'] = 0.1

        return self.__parameters
    
    def apply(self):
        """
        Applies the effect to the provided 'clip' and with the also
        provided parameters needed by this effect.
        """
        return self.__clip.fx(self.__get_moviepy_vfx_effect(), **self.__process_parameters())



from TTS.api import TTS

# TODO: From here (https://github.com/coqui-ai/TTS)
def narrate(text, output_filename = None):
    """
    Generates a narration audio file with the provided 'text' that
    will be stored as 'output_filename' file.

    This method uses a Spanish model so 'text' must be in Spanish.

    This method will take some time to generate the narration.
    """
    if not output_filename:
        return None
    
    # tts_es_fastpitch_multispeaker.nemo
    # These below are the 2 Spanish models that exist
    SPANISH_MODEL_A = 'tts_models/es/mai/tacotron2-DDC'
    SPANISH_MODEL_B = 'tts_models/es/css10/vits'
    tts = TTS(model_name = SPANISH_MODEL_B)
    tts.tts_to_file(text = text, file_path = output_filename)

def narrrate_imitating_voice(text, input_filename = None, output_filename = None):
    """
    Narrates the provided 'text' by imitating the provided 'input_filename'
    audio file (that must be a voice narrating something) and saves the 
    narration as 'output_filename'.

    The 'input_filename' could be an array of audio filenames.

    Language is set 'es' in code by default.

    This method will take time as it will recreate the voice parameters with
    which the narration will be created after that.

    ANNOTATIONS: This method is only copying the way the narration voice 
    talks, but not the own voice. This is not working as expected, as we are
    not cloning voices, we are just imitating the tone. We need another way
    to actually clone the voice as Elevenlabs do.
    """
    if not input_filename:
        return None
    if not output_filename:
        return None

    tts = TTS("tts_models/multilingual/multi-dataset/xtts_v2")
    # This below will use the latest XTTS_v2 (needs to download the model)
    #tts = TTS('xtts')

    # TODO: Implement a way of identifying and storing the voices we create to
    # be able to use again them without recreating them twice.

    # input_filename can be an array of wav files
    # generate speech by cloning a voice using default settings
    tts.tts_to_file(text = text, file_path = output_filename, speaker_wav = input_filename, language = 'es')


from manim import AddTextLetterByLetter, Write, DrawBorderThenFill, ApplyWave, FadeIn
from random import randrange

# TODO: Maybe this method below is unnecessary
def simple_play_animation(self, animation, object, duration: float = 1):
    WAITING = duration * 0.05
    ANIMATION_DURATION = duration - 2 * WAITING

    self.wait(WAITING)
    self.play(animation(object), run_time = ANIMATION_DURATION)
    self.wait(WAITING)

# TODO: I think this method below doesn't fit here
def RandomTextAnimation():
    """
    Returns a random text animation to be applied with text.
    """
    random_animations = [AddTextLetterByLetter, Write, DrawBorderThenFill, ApplyWave, FadeIn]

    return random_animations[randrange(len(random_animations))]


import math
import random

PERLIN_YWRAPB = 4
PERLIN_YWRAP = 1<<PERLIN_YWRAPB
PERLIN_ZWRAPB = 8
PERLIN_ZWRAP = 1<<PERLIN_ZWRAPB
PERLIN_SIZE = 4095

perlin_octaves = 4
perlin_amp_falloff = 0.5

def scaled_cosine(i):
    return 0.5*(1.0-math.cos(i*math.pi))

perlin = None

def noise(x,y=0,z=0):
    global perlin
    if perlin == None:
        perlin = []
        for i in range(0,PERLIN_SIZE+1):
            perlin.append(random.random())
    if x<0:x=-x
    if y<0:y=-y
    if z<0:z=-z
    
    xi,yi,zi = int(x),int(y),int(z)
    xf = x-xi
    yf = y-yi
    zf = z-zi
    rxf = ryf = None
    
    r = 0
    ampl = 0.5
    
    n1 = n2 = n3 = None
    for o in range(0,perlin_octaves):
        of=xi+(yi<<PERLIN_YWRAPB)+(zi<<PERLIN_ZWRAPB)

        rxf = scaled_cosine(xf)
        ryf = scaled_cosine(yf)

        n1  = perlin[of&PERLIN_SIZE]
        n1 += rxf*(perlin[(of+1)&PERLIN_SIZE]-n1)
        n2  = perlin[(of+PERLIN_YWRAP)&PERLIN_SIZE]
        n2 += rxf*(perlin[(of+PERLIN_YWRAP+1)&PERLIN_SIZE]-n2)
        n1 += ryf*(n2-n1)

        of += PERLIN_ZWRAP
        n2  = perlin[of&PERLIN_SIZE]
        n2 += rxf*(perlin[(of+1)&PERLIN_SIZE]-n2)
        n3  = perlin[(of+PERLIN_YWRAP)&PERLIN_SIZE]
        n3 += rxf*(perlin[(of+PERLIN_YWRAP+1)&PERLIN_SIZE]-n3)
        n2 += ryf*(n3-n2)

        n1 += scaled_cosine(zf)*(n2-n1)

        r += n1*ampl
        ampl *= perlin_amp_falloff
        xi<<=1
        xf*=2
        yi<<=1
        yf*=2
        zi<<=1
        zf*=2
        
        if (xf>=1.0): xi+=1; xf-=1
        if (yf>=1.0): yi+=1; yf-=1
        if (zf>=1.0): zi+=1; zf-=1      
    return r
        
def noiseDetail(lod, falloff):
    if lod>0:perlin_octaves=lod
    if falloff>0:perlin_amp_falloff=falloff 
    
    
class LCG():
    def __init__(self):
        self.m = 4294967296.0
        self.a = 1664525.0
        self.c = 1013904223.0   
        self.seed = self.z = None
    def setSeed(self,val=None):
        self.z = self.seed = (math.random()*self.m if val == None else val) >> 0
    def getSeed(self):
        return self.seed
    def rand(self):
        self.z = (self.a * self.z + self.c) % self.m
        return self.z/self.m        
        
    
def noiseSeed(seed):
    lcg = LCG()
    lcg.setSeed(seed)
    perlin = []
    for i in range(0,PERLIN_SIZE+1):
        perlin.append(lcg.rand())
        
        


from moviepy.editor import VideoFileClip, CompositeVideoClip, ImageClip
from typing import Union
from skimage.filters import gaussian

class BlurMoviepyEffect:
    """
    This effect will zoom out the clip, on the center.

    TODO: This effect is not smooth as it makes it have
    a black border. Maybe removing it (?)
    """
    __parameters = {}

    def __init__(self, clip: Union[VideoFileClip, CompositeVideoClip, ImageClip], blur_radius = 4):
        self.__clip = clip
        self.__parameters['blur_radius'] = blur_radius

    def __process_parameters(self):
        if not self.__parameters['blur_radius']:
            self.__parameters['blur_radius'] = 4
        else:
            # Zoom is by now limited to [2 - 64] ratio
            if self.__parameters['blur_radius'] > 64:
                self.__parameters['blur_radius'] = 64
            elif self.__parameters['blur_radius'] <= 2:
                self.__parameters['blur_radius'] = 2

        return self.__parameters
    
    def __effect_calculations(self, get_frame, t, blur_radius):
        return gaussian(get_frame(t).astype(float), sigma = blur_radius)
    
    def apply(self):
        """
        Applies the effect to the provided 'clip' and with the also
        provided parameters needed by this effect.
        """
        return self.__clip.fl(lambda get_frame, t: self.__effect_calculations(get_frame, t, **self.__process_parameters()))



from yta_general_utils.text_processor import remove_non_ascii_characters

import requests
import base64

def narrate_tiktok(text: str, output_filename: str):
    """
    This is the tiktok voice based on a platform that generates it.
    This will make a narration with the tiktok voice. You can
    change the code to use the mexican voice.

    As this is based on an external platform, it could fail.
    """
    # From here: https://gesserit.co/tiktok    
    # A project to use Tiktok API and cookie (https://github.com/Steve0929/tiktok-tts)
    # A project to use Tiktok API and session id (https://github.com/oscie57/tiktok-voice)
    # A project that is install and play (I think) https://github.com/Giooorgiooo/TikTok-Voice-TTS/blob/main/tiktokvoice.py

    headers = {
        'accept': '*/*',
        'accept-language': 'es-ES,es;q=0.9',
        'content-type': 'text/plain;charset=UTF-8',
        'origin': 'https://gesserit.co',
        'referer': 'https://gesserit.co/tiktok',
        'sec-ch-ua': '"Google Chrome";v="123", "Not:A-Brand";v="8", "Chromium";v="123"',
        'sec-ch-ua-mobile': '?0',
        'sec-ch-ua-platform': '"Windows"',
        'sec-fetch-dest': 'empty',
        'sec-fetch-mode': 'cors',
        'sec-fetch-site': 'same-origin',
        'user-agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/123.0.0.0 Safari/537.36',
    }

    # Non-English characters are not accepted by Tiktok TTS generation, so:
    text = remove_non_ascii_characters(text)

    # TODO: There a a lot of English US and more languages voices
    # These voices below are Spanish
    MEXICAN_VOICE = 'es_mx_002'
    SPANISH_VOICE = 'es_002'
    data = '{"text":"' + text + '","voice":"' + SPANISH_VOICE + '"}'

    response = requests.post('https://gesserit.co/api/tiktok-tts', headers=headers, data=data)
    response = response.json()
    base64_content = response['base64']

    if not output_filename.endswith('.mp3'):
        output_filename += '.mp3'

    try:
        content = base64.b64decode(base64_content)
        with open(output_filename,"wb") as f:
            f.write(content)
    except Exception as e:
        print(str(e))

    return output_filename
    


"""
Mobject to represent a video inside of Manim. This video
Mobject is a modified ImageMobject that updates the 
content on each iteration.

This mobject is working properly with 'cairo' render and
using the 'ImageMobject' class for the inheritance.

TODO: Make this work with 'opengl' render by replacing 
the 'ImageMobject' with 'OpenGLImageMobject', but this
is returning a 'astype' exception when formating the
content with PIL lib in '.astype('uint8')'.

Thanks: https://gist.github.com/uwezi/faec101ed5d7c20222b33eee4b6c7d63
"""
from manim import *
from dataclasses import dataclass
from PIL import Image

import cv2

@dataclass
class VideoStatus:
    time: float = 0
    videoObject: cv2.VideoCapture = None

    def __deepcopy__(self, memo):
        return self

# TODO: Try to use 'OpenGLImageMobject' to be able
# to render in a 3D envinroment
# If trying with 'OpenGLImageMobject' you should
# use the 'opengl' animator , but this is failing
class VideoMobject(ImageMobject):
    '''
    Mobject that represents a video as a sequence of ImageMobjects
    that are updated as a video: frame by frame.

    This VideoMobject has been created to work with 'Cairo' render.
    It is working perfectly in 2D scenes and the video is also being
    displayed in 3D scenes but, in those cases, the positioning is
    being crazy, you cannot control the cameras.

    Parameters
    ----------
    filename
        the filename of the video file

    loop
        (optional) replay the video from the start in an endless loop

    https://discord.com/channels/581738731934056449/1126245755607339250/1126245755607339250
    2023-07-06 Uwe Zimmermann & Abulafia
    2024-03-09 Uwe Zimmermann
    '''
    def __init__(self, filename = None, loop = True, **kwargs):
        self.filename = filename
        self.speed = 1
        self.loop = loop
        self._id = id(self)
        self.status = VideoStatus()
        self.status.videoObject = cv2.VideoCapture(filename)

        self.status.videoObject.set(cv2.CAP_PROP_POS_FRAMES, 1)
        ret, frame = self.status.videoObject.read()
        
        if ret:
            frame = cv2.cvtColor(frame, cv2.COLOR_BGR2RGB)            
            img = Image.fromarray(frame)
        else:
            img = Image.fromarray(np.uint8([[63, 0, 0, 0],
                                        [0, 127, 0, 0],
                                        [0, 0, 191, 0],
                                        [0, 0, 0, 255]
                                        ]))
        self.size = img.size

        super().__init__(img, **kwargs)

        if ret:
            self.add_updater(self.video_updater)

    def video_updater(self, mobj, dt):
        """
        This method updates the image (frame) that needs to be
        shown in the animation attending to the moment.
        """
        if dt == 0:
            return
        
        status = self.status
        status.time += 1000 * dt * mobj.speed
        self.status.videoObject.set(cv2.CAP_PROP_POS_MSEC, status.time)
        ret, frame = self.status.videoObject.read()

        # Go to frame 0 if ended and loop requested
        if (ret == False) and self.loop:
            status.time = 0
            self.status.videoObject.set(cv2.CAP_PROP_POS_MSEC, status.time)
            ret, frame = self.status.videoObject.read()

        if ret:
            frame = cv2.cvtColor(frame, cv2.COLOR_BGR2RGB)  # needed here?        
            img = Image.fromarray(frame)

            # Update the content (the image shown)
            mobj.pixel_array = change_to_rgba_array(
                np.asarray(img), mobj.pixel_array_dtype
            )


from moviepy.editor import VideoFileClip, CompositeVideoClip, ImageClip, AudioFileClip, ColorClip, vfx
from typing import Union
from yta_multimedia.resources.video.effect.sound.drive_urls import SAD_MOMENT_GOOGLE_DRIVE_DOWNLOAD_URL
from yta_multimedia.resources import get_resource
from yta_multimedia.video.edition.effect.constants import EFFECTS_RESOURCES_FOLDER

class SadMomentMoviepyEffect:
    """
    This method gets the first frame of the provided 'clip' and returns a
    new clip that is an incredible 'sad_moment' effect with black and white
    filter, zoom in and rotating effect and also sad violin music.

    The 'duration' parameter is to set the returned clip duration, but the
    default value is a perfect one.
    """
    __parameters = {}

    def __init__(self, clip: Union[VideoFileClip, CompositeVideoClip, ImageClip], duration = 4.8):
        self.__clip = clip
        self.__parameters['duration'] = duration

    def __process_parameters(self):
        if not self.__parameters['duration']:
            self.__parameters['duration'] = 4.8
        else:
            # Zoom is by now limited to [1 - 10] ratio
            if self.__parameters['duration'] > 10:
                self.__parameters['duration'] = 10
            elif self.__parameters['duration'] <= 1:
                self.__parameters['duration'] = 1

        return self.__parameters
    
    def apply(self):
        """
        Applies the effect to the provided 'clip' and with the also
        provided parameters needed by this effect.
        """
        if not self.__clip:
            return None
        
        self.__process_parameters()
        
        # We freeze the first frame
        aux = ImageClip(self.__clip.get_frame(0), duration = self.__parameters['duration'])
        aux.fps = self.__clip.fps
        self.__clip = aux
        # We then build the whole effect
        self.__clip = self.__clip.fx(vfx.blackwhite).resize(lambda t: 1 + 0.30 * (t / self.__clip.duration)).set_position(lambda t: (-(0.15 * self.__clip.w * (t / self.__clip.duration)), -(0.15 * self.__clip.h * (t / self.__clip.duration)))).rotate(lambda t: 5 * (t / self.__clip.duration), expand = False)
        # We set the effect audio
        TMP_FILENAME = get_resource(SAD_MOMENT_GOOGLE_DRIVE_DOWNLOAD_URL, EFFECTS_RESOURCES_FOLDER + 'sounds/sad_moment.mp3')
        self.__clip.audio = AudioFileClip(TMP_FILENAME).set_duration(self.__clip.duration)

        return CompositeVideoClip([
            ColorClip(color = [0, 0, 0], size = self.__clip.size, duration = self.__clip.duration),
            self.__clip,
        ])


from typing import Tuple, Union

class GreenscreenAreaDetails:
    """
    This class represents a greenscreen area inside of a greenscren
    video or image resource and will containg the used rgb color, 
    the position and more information.

    @param
    **rgb_color**
        The green color in rgb format: (r, g, b).
    @param
    **upper_left_pixel**
        The upper left pixel position in the format (x, y).
    @param
    **lower_right_pixel**
        The lower right pixel position in the format (x, y).
    @param
    **frames**
        The frames in which the greenscreen area is present in the
        (start, end) format. If the greenscreen area is present the
        whole video or the greenscreen is an image, this value
        will be None.
    
    """
    rgb_color = None
    upper_left_pixel = None
    lower_right_pixel = None
    frames = None

    def __init__(self, rgb_color: Tuple[int, int, int] = (0, 0, 255), upper_left_pixel: Tuple[int, int] = (0, 0), lower_right_pixel: Tuple[int, int] = (0, 0), frames: Union[Tuple[int, int], None] = None):
        # TODO: Implement checkings please
        self.rgb_color = rgb_color
        self.upper_left_pixel = upper_left_pixel
        self.lower_right_pixel = lower_right_pixel
        self.frames = frames